#  Copyright 2025 European Union
#  Author: Bulgheroni Antonio (antonio.bulgheroni@ec.europa.eu)
#  SPDX-License-Identifier: EUPL-1.2
"""
This is the module that will be exposed via the entry point declaration.

Make sure to have all processors that you need to export in the list.
"""

from plug.db_model import Detector
from plug.plug_processor import Analyser, GenerateDataFiles, PlugImporter, PlugPlotter

import mafw


@mafw.mafw_hookimpl
def register_processors() -> list[mafw.processor.Processor]:
    return [GenerateDataFiles, PlugImporter, Analyser, PlugPlotter]


@mafw.mafw_hookimpl
def register_standard_tables() -> list[mafw.db.std_tables.StandardTable]:
    return [Detector]



#  Copyright 2025 European Union
#  Author: Bulgheroni Antonio (antonio.bulgheroni@ec.europa.eu)
#  SPDX-License-Identifier: EUPL-1.2
"""
Module provides standard tables that are included in all database created by MAFw processor.

Standard tables are automatically created and initialized by a :class:`~mafw.processor.Processor` or a
:class:`~mafw.processor.ProcessorList` when opening a database connection.

This means that if a processor receives a valid database object, then it will suppose that the connection was already
opened somewhere else (either from a ProcessorList or a third party) and thus it is not creating the standard tables.

If a processor is constructed using a database configuration dictionary, then it will first try to open a connection
to the DB, then creating all standard tables and finally executing their :class:`StandardTable.init` method. The same
apply for the Processor list.

In other words, object responsible to open the database connection is taking care also of creating the standard
tables and of initializing them. If the user opens the connection and passes it to a Processor or ProcessorList,
then the user is responsible to create the standard tables and to initialize them.

All standard tables must derive from the :class:`StandardTable` to have the same interface for the
initialization.

Users can create their own standard tables and export them as a pluggable object using the same approach as for custom
processors. An example is provided in the :ref:`documentation <my_std_tables.py>`.

"""

from types import TracebackType
from typing import cast

import peewee
from peewee import AutoField, BooleanField, CharField, TextField

from mafw.db.db_model import MAFwBaseModel
from mafw.db.db_types import PeeweeModelWithMeta
from mafw.db.fields import FileChecksumField, FileNameListField
from mafw.db.trigger import Trigger, TriggerAction, TriggerWhen


class StandardTable(MAFwBaseModel):
    """A base class for tables that are generated automatically by the MAFw processor."""

    @classmethod
    def init(cls) -> None:
        """The user must overload this method, if he wants some specific operations to be performed on the model
        everytime the database is connected."""
        pass


class StandardTableDoesNotExist(Exception):
    """An exception raised when trying to access a not existing table."""


class TriggerStatus(StandardTable):
    """A Model for the trigger status"""

    trigger_type_id = AutoField(primary_key=True, help_text='Primary key')
    trigger_type = TextField(
        help_text='You can use it to specify the type (DELETE/INSERT/UPDATE) or the name of a specific trigger'
    )
    status = BooleanField(default=True, help_text='False (0) = disable / True (1) = enable')

    # noinspection PyProtectedMember
    @classmethod
    def init(cls) -> None:
        """Resets all triggers to enable when the database connection is opened."""
        data = [
            dict(trigger_type_id=1, trigger_type='DELETE', status=True),
            dict(trigger_type_id=2, trigger_type='INSERT', status=True),
            dict(trigger_type_id=3, trigger_type='UPDATE', status=True),
            dict(trigger_type_id=4, trigger_type='DELETE_FILES', status=True),
        ]

        # this is used just to make mypy happy
        # cls and meta_cls are exactly the same thing
        meta_cls = cast(PeeweeModelWithMeta, cls)

        db_proxy = meta_cls._meta.database
        if isinstance(db_proxy, peewee.DatabaseProxy):
            db = cast(peewee.Database, db_proxy.obj)
        else:
            db = cast(peewee.Database, db_proxy)

        if isinstance(db, peewee.PostgresqlDatabase):
            cls.insert_many(data).on_conflict(
                'update', conflict_target=[cls.trigger_type_id], update={cls.status: True}
            ).execute()
        else:
            cls.insert_many(data).on_conflict_replace().execute()


class TriggerStatusDoesNotExist(Exception):
    """An exception raised when trying to access a not existing table."""


class TriggerDisabler:
    """
    A helper tool to disable a specific type of triggers.

    Not all SQL dialects allow to temporarily disable trigger execution.

    In order overcome this limitation, MAFw has introduced a practical workaround. All types of triggers are active
    by default but they can be temporarily disabled, by changing their status in the :class:`.TriggerStatus` table.

    In order to disable the trigger execution, the user has to set the status of the corresponding status to 0 and also
    add a when condition to the trigger definition.

    Here is an example code:

    .. code-block:: python

        class MyTable(MAFwBaseModel):
            id_ = AutoField(primary_key=True)
            integer = IntegerField()
            float_num = FloatField()

            @classmethod
            def triggers(cls):
                return [
                    Trigger(
                        'mytable_after_insert',
                        (TriggerWhen.After, TriggerAction.Insert),
                        cls,
                        safe=True,
                    )
                    .add_sql(
                        'INSERT INTO target_table (id__id, half_float_num) VALUES (NEW.id_, NEW.float_num / 2)'
                    )
                    .add_when(
                        '1 == (SELECT status FROM trigger_status WHERE trigger_type_id == 1)'
                    )
                ]

    When you want to perform a database action with the trigger disabled, you can either use this class as context
    manager or call the :meth:`.disable` and :meth:`.enable` methods.

    .. code-block:: python

        # as a context manager
        with TriggerDisabler(trigger_type_id = 1):
            # do something without triggering any trigger of type 1.

        # with the explicit methods
        disabler = TriggerDisabler(1)
        disabler.disable()
        # do something without triggering any trigger of type 1.
        disabler.enable()

    When using the two explicit methods, the responsibility to assure that the triggers are re-enabled in on the user.
    """

    def __init__(self, trigger_type_id: int) -> None:
        """
        Constructor parameters:

        :param trigger_type_id: the id of the trigger to be temporary disabled.
        :type trigger_type_id: int
        """
        self.trigger_type_id = trigger_type_id

    def disable(self) -> None:
        """Disable the trigger"""
        TriggerStatus.update({TriggerStatus.status: 0}).where(
            TriggerStatus.trigger_type_id == self.trigger_type_id
        ).execute()

    def enable(self) -> None:
        """Enable the trigger"""
        TriggerStatus.update({TriggerStatus.status: 1}).where(
            TriggerStatus.trigger_type_id == self.trigger_type_id
        ).execute()

    def __enter__(self) -> 'TriggerDisabler':
        """
        Context enter. Disable the trigger.
        """
        self.disable()
        return self

    def __exit__(
        self, type_: type[BaseException] | None, value: BaseException | None, traceback: TracebackType | None
    ) -> None:
        """
        Context exit

        :param type_: Exception type causing the context manager to exit. Defaults to None.
        :type type_: type[BaseException], Optional
        :param value: Exception that caused the context manager to exit. Defaults to None.
        :type value: BaseException, Optional
        :param traceback: Traceback. Defaults to None.
        :type traceback: TracebackType
        """
        self.enable()


class OrphanFile(StandardTable):
    """A Model for the files to be removed from disc"""

    file_id = AutoField(primary_key=True, help_text='Primary key')
    filenames = FileNameListField(help_text='The path to the file to be deleted', checksum_field='checksum')
    checksum = FileChecksumField(help_text='The checksum of the files in the list.')


class OrphanFileDoesNotExist(peewee.DoesNotExist):
    """An exception raised when trying to access a not existing table."""


class PlotterOutput(StandardTable):
    """
    A model for the output of the plotter processors.

    The model has a trigger activated on delete queries to insert filenames and checksum in the OrphanFile model.
    """

    plotter_name = CharField(primary_key=True, help_text='The plotter processor name', max_length=511)
    filename_list = FileNameListField(help_text='The path to the output file', checksum_field='checksum')
    checksum = FileChecksumField(help_text='The checksum of the files in the list.')

    @classmethod
    def triggers(cls) -> list[Trigger]:
        insert_into_orphan = Trigger('plotter_delete_file', (TriggerWhen.Before, TriggerAction.Delete), cls, safe=True)
        # warning:
        #   we cannot use the filenames for both the filenames and checksum fields because it is a bare sql statement
        insert_into_orphan.add_sql(
            'INSERT INTO orphan_file (filenames, checksum) VALUES (OLD.filename_list, OLD.checksum);'
        )
        insert_into_orphan.add_when('1 == (SELECT status FROM trigger_status WHERE trigger_type_id == 4)')
        return [insert_into_orphan]

    class Meta:
        depends_on = [OrphanFile]


class PlotterOutputDoesNotExist(peewee.DoesNotExist):
    """An exception raised when trying to access a not existing table."""


standard_tables: dict[str, type[StandardTable]] = {}
"""
A dictionary containing the standard tables being exported to the rest of the framework. 

The key is the name of the model and the value is the model class itself.
"""

standard_tables.update({'TriggerStatus': TriggerStatus, 'OrphanFile': OrphanFile, 'PlotterOutput': PlotterOutput})



#  Copyright 2025 European Union
#  Author: Bulgheroni Antonio (antonio.bulgheroni@ec.europa.eu)
#  SPDX-License-Identifier: EUPL-1.2
"""
Unit tests for trigger.py module.

Tests for and_, or_, TriggerWhen, TriggerAction, and Trigger classes.
"""

import sys
from unittest.mock import Mock, patch

import peewee
import pytest
from peewee import Model

from mafw.db.trigger import (
    Trigger,
    TriggerAction,
    TriggerWhen,
    and_,
    or_,
)
from mafw.mafw_errors import MissingSQLStatement, UnsupportedDatabaseError


class TestLogicalOperators:
    """Test cases for and_ and or_ functions."""

    def test_and_single_condition(self):
        """Test and_ function with single condition."""
        result = and_('condition1')
        assert result == '(condition1)'

    def test_and_multiple_conditions(self):
        """Test and_ function with multiple conditions."""
        result = and_('condition1', 'condition2', 'condition3')
        assert result == '(condition1) AND (condition2) AND (condition3)'

    def test_and_empty_conditions(self):
        """Test and_ function with no conditions."""
        result = and_()
        assert result == ''

    def test_or_single_condition(self):
        """Test or_ function with single condition."""
        result = or_('condition1')
        assert result == '(condition1)'

    def test_or_multiple_conditions(self):
        """Test or_ function with multiple conditions."""
        result = or_('condition1', 'condition2', 'condition3')
        assert result == '(condition1) OR (condition2) OR (condition3)'

    def test_or_empty_conditions(self):
        """Test or_ function with no conditions."""
        result = or_()
        assert result == ''

    def test_and_with_complex_conditions(self):
        """Test and_ function with complex conditions containing spaces and operators."""
        result = and_("field1 = 'value1'", 'field2 > 10', 'field3 IS NOT NULL')
        expected = "(field1 = 'value1') AND (field2 > 10) AND (field3 IS NOT NULL)"
        assert result == expected

    def test_or_with_complex_conditions(self):
        """Test or_ function with complex conditions containing spaces and operators."""
        result = or_("field1 = 'value1'", 'field2 > 10', 'field3 IS NOT NULL')
        expected = "(field1 = 'value1') OR (field2 > 10) OR (field3 IS NOT NULL)"
        assert result == expected


class TestTriggerWhen:
    """Test cases for TriggerWhen enum."""

    def test_trigger_when_values(self):
        """Test that TriggerWhen enum has correct values."""
        assert TriggerWhen.Before == 'BEFORE'
        assert TriggerWhen.After == 'AFTER'
        assert TriggerWhen.Instead == 'INSTEAD OF'

    def test_trigger_when_string_representation(self):
        """Test string representation of TriggerWhen enum values."""
        assert str(TriggerWhen.Before) == 'BEFORE'
        assert str(TriggerWhen.After) == 'AFTER'
        assert str(TriggerWhen.Instead) == 'INSTEAD OF'

    @pytest.mark.skipif(sys.version_info < (3, 12), reason='bug in python < 3.12')
    def test_trigger_when_enum_membership(self):
        """Test membership in TriggerWhen enum."""
        assert 'BEFORE' in TriggerWhen
        assert 'AFTER' in TriggerWhen
        assert 'INSTEAD OF' in TriggerWhen
        assert 'INVALID' not in TriggerWhen


class TestTriggerAction:
    """Test cases for TriggerAction enum."""

    def test_trigger_action_values(self):
        """Test that TriggerAction enum has correct values."""
        assert TriggerAction.Delete == 'DELETE'
        assert TriggerAction.Insert == 'INSERT'
        assert TriggerAction.Update == 'UPDATE'

    def test_trigger_action_string_representation(self):
        """Test string representation of TriggerAction enum values."""
        assert str(TriggerAction.Delete) == 'DELETE'
        assert str(TriggerAction.Insert) == 'INSERT'
        assert str(TriggerAction.Update) == 'UPDATE'

    @pytest.mark.skipif(sys.version_info < (3, 12), reason='bug in python < 3.12')
    def test_trigger_action_enum_membership(self):
        """Test membership in TriggerAction enum."""
        assert 'DELETE' in TriggerAction
        assert 'INSERT' in TriggerAction
        assert 'UPDATE' in TriggerAction
        assert 'INVALID' not in TriggerAction


class TestTrigger:
    """Test cases for Trigger class."""

    @pytest.fixture
    def mock_model_class(self):
        """Mock model class with _meta attribute."""
        mock_class = Mock(spec=type)
        mock_class._meta = Mock()
        mock_class._meta.table_name = 'test_table'
        return mock_class

    @pytest.fixture
    def mock_model_instance(self):
        """Mock model instance with _meta attribute."""
        mock_instance = Mock(spec=Model)
        mock_instance._meta = Mock()
        mock_instance._meta.table_name = 'test_table'
        return mock_instance

    @pytest.fixture
    def mock_database(self):
        """Mock database instance."""
        return Mock(spec=peewee.SqliteDatabase)

    @pytest.fixture
    def mock_database_proxy(self):
        """Mock database proxy instance."""
        proxy = Mock(spec=peewee.DatabaseProxy)
        proxy.obj = Mock(spec=peewee.SqliteDatabase)
        return proxy

    @pytest.fixture
    def basic_trigger(self):
        """Basic trigger instance for testing."""
        return Trigger(
            trigger_name='test_trigger',
            trigger_type=(TriggerWhen.After, TriggerAction.Insert),
            source_table='test_table',
        )

    def test_trigger_init_with_string_table(self):
        """Test Trigger initialization with string table name."""
        trigger = Trigger(
            trigger_name='test_trigger',
            trigger_type=(TriggerWhen.After, TriggerAction.Insert),
            source_table='test_table',
        )

        assert trigger.trigger_name == 'test_trigger'
        assert trigger.trigger_when == TriggerWhen.After
        assert trigger.trigger_action == TriggerAction.Insert
        assert trigger.target_table == 'test_table'
        assert trigger.safe is False
        assert trigger.for_each_row is False
        assert trigger.update_columns == []
        assert trigger.on_view is False

    def test_trigger_init_with_model_class(self, mock_model_class):
        """Test Trigger initialization with model class."""
        trigger = Trigger(
            trigger_name='test_trigger',
            trigger_type=(TriggerWhen.Before, TriggerAction.Update),
            source_table=mock_model_class,
        )

        assert trigger.target_table == 'test_table'

    def test_trigger_init_with_model_instance(self, mock_model_instance):
        """Test Trigger initialization with model instance."""
        trigger = Trigger(
            trigger_name='test_trigger',
            trigger_type=(TriggerWhen.Instead, TriggerAction.Delete),
            source_table=mock_model_instance,
        )

        assert trigger.target_table == 'test_table'

    def test_trigger_init_with_optional_parameters(self):
        """Test Trigger initialization with all optional parameters."""
        trigger = Trigger(
            trigger_name='test_trigger',
            trigger_type=(TriggerWhen.Before, TriggerAction.Update),
            source_table='test_table',
            safe=True,
            for_each_row=True,
            update_columns=['col1', 'col2'],
            on_view=True,
        )

        assert trigger.safe is True
        assert trigger.for_each_row is True
        assert trigger.update_columns == ['col1', 'col2']
        assert trigger.on_view is True

    def test_trigger_property_setters(self, basic_trigger):
        """Test trigger action and when property setters."""
        basic_trigger.trigger_action = TriggerAction.Update
        basic_trigger.trigger_when = TriggerWhen.Before

        assert basic_trigger.trigger_action == TriggerAction.Update
        assert basic_trigger.trigger_when == TriggerWhen.Before

    def test_setattr_safe_property(self, basic_trigger):
        """Test __setattr__ for safe property."""
        basic_trigger.safe = True
        assert hasattr(basic_trigger, 'if_not_exists')
        assert basic_trigger.if_not_exists == 'IF NOT EXISTS'

        basic_trigger.safe = False
        assert basic_trigger.if_not_exists == ''

    def test_setattr_for_each_row_property(self, basic_trigger):
        """Test __setattr__ for for_each_row property."""
        basic_trigger.for_each_row = True
        assert hasattr(basic_trigger, '_for_each_row')
        assert basic_trigger._for_each_row == 'FOR EACH ROW'

        basic_trigger.for_each_row = False
        assert basic_trigger._for_each_row == ''

    def test_getattr_safe_property(self, basic_trigger):
        """Test __getattr__ for safe property."""
        basic_trigger.if_not_exists = 'IF NOT EXISTS'
        result = basic_trigger.__getattr__('safe')
        assert result is True

        basic_trigger.if_not_exists = ''
        result = basic_trigger.__getattr__('safe')
        assert result is False

    def test_getattr_for_each_row_property(self, basic_trigger):
        """Test __getattr__ for for_each_row property."""
        basic_trigger._for_each_row = 'FOR EACH ROW'
        result = basic_trigger.__getattr__('for_each_row')
        assert result is True

        basic_trigger._for_each_row = ''
        result = basic_trigger.__getattr__('for_each_row')
        assert result is False

    def test_getattr_nonexistent_attribute(self, basic_trigger):
        """Test __getattr__ for non-existent attribute raises AttributeError."""
        with pytest.raises(AttributeError, match="'Trigger' object has no attribute 'nonexistent'"):
            basic_trigger.__getattr__('nonexistent')

    def test_getattr_safe_property_without_if_not_exists(self, basic_trigger):
        """Test __getattr__ for safe property when if_not_exists doesn't exist."""
        # Don't set if_not_exists at all
        result = basic_trigger.__getattr__('safe')
        assert result is False

    def test_getattr_for_each_row_without_internal_attr(self, basic_trigger):
        """Test __getattr__ for for_each_row property when _for_each_row doesn't exist."""
        # Don't set _for_each_row at all
        result = basic_trigger.__getattr__('for_each_row')
        assert result is False

    def test_add_sql_with_string(self, basic_trigger):
        """Test add_sql method with string input."""
        sql = "INSERT INTO another_table VALUES (1, 'test')"
        result = basic_trigger.add_sql(sql)

        assert result is basic_trigger  # Method should return self for chaining
        assert len(basic_trigger._sql_list) == 1
        assert basic_trigger._sql_list[0] == "\tINSERT INTO another_table VALUES (1, 'test');"

    def test_add_sql_with_query_object(self, basic_trigger):
        """Test add_sql method with peewee Query object."""
        mock_query = Mock()
        mock_query.__str__ = Mock(return_value='SELECT * FROM test')

        result = basic_trigger.add_sql(mock_query)

        assert result is basic_trigger
        assert len(basic_trigger._sql_list) == 1
        assert basic_trigger._sql_list[0] == '\tSELECT * FROM test;'

    def test_add_sql_with_semicolon(self, basic_trigger):
        """Test add_sql method with SQL that already has semicolon."""
        sql = 'DELETE FROM test_table WHERE id = 1;'
        basic_trigger.add_sql(sql)

        assert basic_trigger._sql_list[0] == '\tDELETE FROM test_table WHERE id = 1;'

    def test_add_sql_multiple_statements(self, basic_trigger):
        """Test adding multiple SQL statements."""
        basic_trigger.add_sql('INSERT INTO table1 VALUES (1)')
        basic_trigger.add_sql("UPDATE table2 SET col1 = 'value'")

        assert len(basic_trigger._sql_list) == 2
        assert basic_trigger._sql_list[0] == '\tINSERT INTO table1 VALUES (1);'
        assert basic_trigger._sql_list[1] == "\tUPDATE table2 SET col1 = 'value';"

    def test_add_when_single_condition(self, basic_trigger):
        """Test add_when method with single condition."""
        result = basic_trigger.add_when('NEW.field1 > OLD.field1')

        assert result is basic_trigger
        assert len(basic_trigger._when_list) == 1
        assert basic_trigger._when_list[0] == '((NEW.field1 > OLD.field1))'

    def test_add_when_multiple_conditions(self, basic_trigger):
        """Test add_when method with multiple conditions."""
        basic_trigger.add_when('NEW.field1 > OLD.field1', 'NEW.field2 IS NOT NULL')

        assert len(basic_trigger._when_list) == 1
        assert basic_trigger._when_list[0] == '((NEW.field1 > OLD.field1) AND (NEW.field2 IS NOT NULL))'

    def test_add_when_multiple_calls(self, basic_trigger):
        """Test multiple calls to add_when method."""
        basic_trigger.add_when('condition1')
        basic_trigger.add_when('condition2', 'condition3')

        assert len(basic_trigger._when_list) == 2
        assert basic_trigger._when_list[0] == '((condition1))'
        assert basic_trigger._when_list[1] == '((condition2) AND (condition3))'

    def test_set_database(self, basic_trigger, mock_database):
        """Test set_database method."""
        result = basic_trigger.set_database(mock_database)

        assert result is basic_trigger
        assert basic_trigger._database is mock_database

    @patch('mafw.db.trigger.SQLiteDialect')
    def test_get_dialect_sqlite(self, mock_sqlite_dialect, basic_trigger, mock_database):
        """Test _get_dialect method with SQLite database."""
        mock_sqlite_db = Mock(spec=peewee.SqliteDatabase)
        basic_trigger._database = mock_sqlite_db

        basic_trigger._get_dialect()

        mock_sqlite_dialect.assert_called_once()
        assert basic_trigger._dialect is not None

    @patch('mafw.db.trigger.MySQLDialect')
    def test_get_dialect_mysql(self, mock_mysql_dialect, basic_trigger):
        """Test _get_dialect method with MySQL database."""
        mock_mysql_db = Mock(spec=peewee.MySQLDatabase)
        basic_trigger._database = mock_mysql_db

        basic_trigger._get_dialect()

        mock_mysql_dialect.assert_called_once()

    @patch('mafw.db.trigger.PostgreSQLDialect')
    def test_get_dialect_postgresql(self, mock_postgresql_dialect, basic_trigger):
        """Test _get_dialect method with PostgreSQL database."""
        mock_postgresql_db = Mock(spec=peewee.PostgresqlDatabase)
        basic_trigger._database = mock_postgresql_db

        basic_trigger._get_dialect()

        mock_postgresql_dialect.assert_called_once()

    def test_get_dialect_database_proxy(self, basic_trigger, mock_database_proxy):
        """Test _get_dialect method with database proxy."""
        basic_trigger._database = mock_database_proxy

        with patch('mafw.db.trigger.SQLiteDialect') as mock_sqlite_dialect:
            basic_trigger._get_dialect()
            mock_sqlite_dialect.assert_called_once()

    def test_get_dialect_unsupported_database(self, basic_trigger):
        """Test _get_dialect method with unsupported database."""
        unsupported_db = Mock()
        basic_trigger._database = unsupported_db

        with pytest.raises(UnsupportedDatabaseError):
            basic_trigger._get_dialect()

    def test_get_dialect_no_database_defaults_to_sqlite(self, basic_trigger):
        """Test _get_dialect method with no database set (defaults to SQLite)."""
        with patch('mafw.db.trigger.SQLiteDialect') as mock_sqlite_dialect:
            basic_trigger._get_dialect()
            mock_sqlite_dialect.assert_called_once()

    def test_get_dialect_caching(self, basic_trigger, mock_database):
        """Test that _get_dialect caches the dialect instance."""
        basic_trigger._database = mock_database

        with patch('mafw.db.trigger.SQLiteDialect') as mock_sqlite_dialect:
            dialect1 = basic_trigger._get_dialect()
            dialect2 = basic_trigger._get_dialect()

            # Should only be called once due to caching
            mock_sqlite_dialect.assert_called_once()
            assert dialect1 is dialect2

    def test_create_no_sql_statements(self, basic_trigger):
        """Test create method with no SQL statements raises error."""
        with pytest.raises(MissingSQLStatement):
            basic_trigger.create()

    def test_create_with_sql_statements(self, basic_trigger):
        """Test create method with SQL statements."""
        basic_trigger.add_sql('INSERT INTO test VALUES (1)')

        with patch.object(basic_trigger, '_get_dialect') as mock_get_dialect:
            mock_dialect = Mock()
            mock_dialect.supports_trigger_type.return_value = True
            mock_dialect.supports_safe_create.return_value = True
            mock_dialect.supports_update_of_columns.return_value = True
            mock_dialect.create_trigger_sql.return_value = 'CREATE TRIGGER test_sql'
            mock_get_dialect.return_value = mock_dialect

            result = basic_trigger.create()

            assert result == 'CREATE TRIGGER test_sql'
            mock_dialect.supports_trigger_type.assert_called_once_with(
                basic_trigger.trigger_when, basic_trigger.trigger_action, basic_trigger.on_view
            )
            mock_dialect.create_trigger_sql.assert_called_once_with(basic_trigger)

    def test_create_unsupported_trigger_type(self, basic_trigger):
        """Test create method with unsupported trigger type."""
        basic_trigger.add_sql('INSERT INTO test VALUES (1)')

        with patch.object(basic_trigger, '_get_dialect') as mock_get_dialect:
            mock_dialect = Mock()
            mock_dialect.supports_trigger_type.return_value = False
            mock_get_dialect.return_value = mock_dialect

            with pytest.raises(UnsupportedDatabaseError):
                basic_trigger.create()

    def test_create_unsupported_safe_create(self, basic_trigger):
        """Test create method when safe create is not supported."""
        basic_trigger.add_sql('INSERT INTO test VALUES (1)')
        basic_trigger.safe = True

        with patch.object(basic_trigger, '_get_dialect') as mock_get_dialect:
            mock_dialect = Mock()
            mock_dialect.supports_trigger_type.return_value = True
            mock_dialect.supports_safe_create.return_value = False
            mock_dialect.supports_update_of_columns.return_value = True
            mock_dialect.create_trigger_sql.return_value = 'CREATE TRIGGER test_sql'
            mock_get_dialect.return_value = mock_dialect

            result = basic_trigger.create()

            assert basic_trigger.safe is False
            assert result == 'CREATE TRIGGER test_sql'

    def test_create_unsupported_update_columns(self, basic_trigger):
        """Test create method when update columns are not supported."""
        basic_trigger.add_sql('INSERT INTO test VALUES (1)')
        basic_trigger.update_columns = ['col1', 'col2']

        with patch.object(basic_trigger, '_get_dialect') as mock_get_dialect:
            mock_dialect = Mock()
            mock_dialect.supports_trigger_type.return_value = True
            mock_dialect.supports_safe_create.return_value = True
            mock_dialect.supports_update_of_columns.return_value = False
            mock_dialect.create_trigger_sql.return_value = 'CREATE TRIGGER test_sql'
            mock_get_dialect.return_value = mock_dialect

            result = basic_trigger.create()

            assert basic_trigger.update_columns == []
            assert result == 'CREATE TRIGGER test_sql'

    def test_drop_default_safe(self, basic_trigger):
        """Test drop method with default safe parameter."""
        with patch.object(basic_trigger, '_get_dialect') as mock_get_dialect:
            mock_dialect = Mock()
            mock_dialect.drop_trigger_sql.return_value = 'DROP TRIGGER IF EXISTS test_trigger'
            mock_get_dialect.return_value = mock_dialect

            result = basic_trigger.drop()

            assert result == 'DROP TRIGGER IF EXISTS test_trigger'
            mock_dialect.drop_trigger_sql.assert_called_once_with('test_trigger', True)

    def test_drop_safe_false(self, basic_trigger):
        """Test drop method with safe=False."""
        with patch.object(basic_trigger, '_get_dialect') as mock_get_dialect:
            mock_dialect = Mock()
            mock_dialect.drop_trigger_sql.return_value = 'DROP TRIGGER test_trigger'
            mock_get_dialect.return_value = mock_dialect

            result = basic_trigger.drop(safe=False)

            assert result == 'DROP TRIGGER test_trigger'
            mock_dialect.drop_trigger_sql.assert_called_once_with('test_trigger', False)

    @pytest.mark.parametrize(
        'trigger_when,trigger_action,expected_when,expected_action',
        [
            (TriggerWhen.Before, TriggerAction.Insert, TriggerWhen.Before, TriggerAction.Insert),
            (TriggerWhen.After, TriggerAction.Update, TriggerWhen.After, TriggerAction.Update),
            (TriggerWhen.Instead, TriggerAction.Delete, TriggerWhen.Instead, TriggerAction.Delete),
        ],
    )
    def test_trigger_types_parametrized(self, trigger_when, trigger_action, expected_when, expected_action):
        """Parametrized test for different trigger types."""
        trigger = Trigger(
            trigger_name='test_trigger', trigger_type=(trigger_when, trigger_action), source_table='test_table'
        )

        assert trigger.trigger_when == expected_when
        assert trigger.trigger_action == expected_action

    @pytest.mark.parametrize(
        'safe,for_each_row,on_view',
        [
            (True, True, True),
            (True, False, False),
            (False, True, False),
            (False, False, True),
        ],
    )
    def test_trigger_boolean_flags_parametrized(self, safe, for_each_row, on_view):
        """Parametrized test for boolean flags."""
        trigger = Trigger(
            trigger_name='test_trigger',
            trigger_type=(TriggerWhen.After, TriggerAction.Insert),
            source_table='test_table',
            safe=safe,
            for_each_row=for_each_row,
            on_view=on_view,
        )

        assert trigger.safe == safe
        assert trigger.for_each_row == for_each_row
        assert trigger.on_view == on_view

    def test_trigger_chaining_methods(self, basic_trigger):
        """Test that methods return self for method chaining."""
        result = basic_trigger.add_sql('INSERT INTO test VALUES (1)').add_when('NEW.id > 0').set_database(Mock())

        assert result is basic_trigger
        assert len(basic_trigger._sql_list) == 1
        assert len(basic_trigger._when_list) == 1
        assert basic_trigger._database is not None



#  Copyright 2025 European Union
#  Author: Bulgheroni Antonio (antonio.bulgheroni@ec.europa.eu)
#  SPDX-License-Identifier: EUPL-1.2
"""
Unit tests for regexp module.
"""

from mafw.tools.regexp import extract_protocol, normalize_sql_spaces


class TestExtractProtocol:
    """Test cases for extract_protocol function."""

    def test_extract_protocol_valid_urls(self):
        """Test extracting protocol from valid URLs."""
        # Common database protocols
        assert extract_protocol('postgresql://user:pass@host:5432/db') == 'postgresql'
        assert extract_protocol('mysql://user:pass@host:3306/db') == 'mysql'
        assert extract_protocol('sqlite:///path/to/database.db') == 'sqlite'
        assert extract_protocol('mongodb://user:pass@host:27017/db') == 'mongodb'

        # Web protocols
        assert extract_protocol('http://example.com') == 'http'
        assert extract_protocol('https://example.com') == 'https'
        assert extract_protocol('ftp://example.com') == 'ftp'

        # Protocols with numbers, hyphens, underscores, dots, and plus signs
        assert extract_protocol('redis+sentinel://host:port') == 'redis+sentinel'
        assert extract_protocol('mssql+pyodbc://server/database') == 'mssql+pyodbc'
        assert extract_protocol('oracle+cx_oracle://user:pass@host') == 'oracle+cx_oracle'
        assert extract_protocol('protocol_123://host') == 'protocol_123'
        assert extract_protocol('my-protocol://host') == 'my-protocol'
        assert extract_protocol('protocol.v2://host') == 'protocol.v2'

    def test_extract_protocol_edge_cases(self):
        """Test edge cases for protocol extraction."""
        # Single character protocol
        assert extract_protocol('a://host') == 'a'

        # Protocol with all allowed characters
        assert extract_protocol('abc123-_+.def://host') == 'abc123-_+.def'

        # Minimal valid URL
        assert extract_protocol('x://') == 'x'

    def test_extract_protocol_invalid_urls(self):
        """Test protocol extraction with invalid URLs."""
        # No protocol separator
        assert extract_protocol('postgresql') is None
        assert extract_protocol('example.com') is None

        # Invalid protocol characters
        assert extract_protocol('POSTGRESQL://host') is None  # uppercase
        assert extract_protocol('post@gres://host') is None  # @ symbol
        assert extract_protocol('post gres://host') is None  # space
        assert extract_protocol('post#gres://host') is None  # # symbol

        # Empty or whitespace
        assert extract_protocol('') is None
        assert extract_protocol('   ') is None

        # Only separator
        assert extract_protocol('://') is None
        assert extract_protocol('://host') is None

        # Protocol doesn't start at beginning
        assert extract_protocol('prefix postgresql://host') is None

    def test_extract_protocol_special_cases(self):
        """Test special cases and potential edge scenarios."""
        # Multiple :// in URL (should match first)
        assert extract_protocol('http://example.com://path') == 'http'

        # URL with port and complex path
        assert extract_protocol('postgresql://user:pass@localhost:5432/mydb?ssl=true') == 'postgresql'

        # Very long protocol name
        long_protocol = 'a' * 50
        assert extract_protocol(f'{long_protocol}://host') == long_protocol


class TestNormalizeSqlSpaces:
    """Test cases for normalize_sql_spaces function."""

    def test_normalize_multiple_spaces(self):
        """Test normalization of multiple consecutive spaces."""
        # Multiple spaces between words
        assert normalize_sql_spaces('SELECT  *  FROM  table') == 'SELECT * FROM table'
        assert normalize_sql_spaces('SELECT   *   FROM   table') == 'SELECT * FROM table'

        # Many consecutive spaces
        assert normalize_sql_spaces('SELECT          *          FROM          table') == 'SELECT * FROM table'

        # Mixed multiple spaces
        assert normalize_sql_spaces('SELECT  *   FROM    table  WHERE     id = 1') == 'SELECT * FROM table WHERE id = 1'

    def test_normalize_leading_trailing_spaces(self):
        """Test trimming of leading and trailing spaces."""
        # Leading spaces
        assert normalize_sql_spaces('  SELECT * FROM table') == 'SELECT * FROM table'
        assert normalize_sql_spaces('     SELECT * FROM table') == 'SELECT * FROM table'

        # Trailing spaces
        assert normalize_sql_spaces('SELECT * FROM table  ') == 'SELECT * FROM table'
        assert normalize_sql_spaces('SELECT * FROM table     ') == 'SELECT * FROM table'

        # Both leading and trailing
        assert normalize_sql_spaces('  SELECT * FROM table  ') == 'SELECT * FROM table'
        assert normalize_sql_spaces('    SELECT * FROM table    ') == 'SELECT * FROM table'

    def test_normalize_preserves_other_whitespace(self):
        """Test that other whitespace characters are preserved."""
        # Tabs should be preserved
        assert normalize_sql_spaces('SELECT\t*\tFROM\ttable') == 'SELECT\t*\tFROM\ttable'

        # Newlines should be preserved
        assert normalize_sql_spaces('SELECT *\nFROM table') == 'SELECT *\nFROM table'

        # Carriage returns should be preserved
        assert normalize_sql_spaces('SELECT *\rFROM table') == 'SELECT *\rFROM table'

        # Mixed whitespace with multiple spaces
        assert (
            normalize_sql_spaces('SELECT  *\nFROM   table\t\tWHERE  id = 1') == 'SELECT *\nFROM table\t\tWHERE id = 1'
        )

    def test_normalize_edge_cases(self):
        """Test edge cases for SQL space normalization."""
        # Empty string
        assert normalize_sql_spaces('') == ''

        # Only spaces
        assert normalize_sql_spaces('   ') == ''
        assert normalize_sql_spaces('  ') == ''

        # Single space
        assert normalize_sql_spaces(' ') == ''

        # String with no spaces
        assert normalize_sql_spaces('SELECT*FROM*table') == 'SELECT*FROM*table'

        # Single word with leading/trailing spaces
        assert normalize_sql_spaces('  word  ') == 'word'

    def test_normalize_complex_sql_examples(self):
        """Test with realistic SQL query examples."""

        # INSERT statement
        sql_input = "INSERT   INTO   users   (name,  email)   VALUES   ('John',  'john@example.com')"
        expected = "INSERT INTO users (name, email) VALUES ('John', 'john@example.com')"
        assert normalize_sql_spaces(sql_input) == expected

        # UPDATE statement
        sql_input = "UPDATE   users   SET   name  =  'Jane'   WHERE   id  =  1  "
        expected = "UPDATE users SET name = 'Jane' WHERE id = 1"
        assert normalize_sql_spaces(sql_input) == expected

    def test_normalize_single_spaces_unchanged(self):
        """Test that strings with only single spaces remain unchanged."""
        # Already normalized strings
        assert normalize_sql_spaces('SELECT * FROM table') == 'SELECT * FROM table'
        assert (
            normalize_sql_spaces("INSERT INTO users (name) VALUES ('John')")
            == "INSERT INTO users (name) VALUES ('John')"
        )
        assert normalize_sql_spaces('a b c d e') == 'a b c d e'



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

import os
import sys

sys.path.insert(0, os.path.abspath('../../src'))
sys.path.insert(0, os.path.abspath('../../src/mafw'))
sys.path.insert(0, os.path.abspath('../../src/mafw/examples'))
import mafw.examples

project = 'MAFw'
copyright = '2024, Antonio Bulgheroni'
author = 'Antonio Bulgheroni'
release = mafw.__about__.__version__

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    'sphinx.ext.duration',
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx_copybutton',
    'sphinx.ext.viewcode',
    'sphinx.ext.extlinks',
    'sphinx.ext.todo',
    'sphinx.ext.doctest',
    'sphinx.ext.graphviz',
    'sphinx_click',
    'sphinx_design',
    'sphinxcontrib.external_links',
    'sphinx.ext.imgconverter',
]

numfig = True
nitpicky = True
nitpick_ignore = [
    ('py:class', 'abc.ABC'),
    ('py:class', 'enum.IntEnum'),
    ('py:class', 'enum.StrEnum'),
    ('py:class', 'enum.Enum'),
    ('py:class', 'pluggy._manager.PluginManager'),
    ('py:class', 'pluggy.PluginManager'),
    ('py:class', 'PluginManager'),
    ('py:class', 'pathlib.Path'),
    ('py:class', 'Path'),
    ('py:exc', 'FileNotFound'),
    ('py:class', 'peewee.Model'),
    ('py:class', 'peewee.ModelBase'),
    ('py:class', 'peewee.Database'),
    ('py:class', 'peewee.DatabaseProxy'),
    ('py:class', 'peewee.TextField'),
    ('py:class', 'peewee.Expression'),
    ('py:class', 'peewee.FieldAccessor'),
    ('py:class', 'peewee.DoesNotExist'),
    ('py:class', 'peewee.ModelInsert'),
    ('py:class', 'peewee.Query'),
    ('py:class', 'playhouse.reflection.Introspector'),
    ('py:class', 'Introspector'),
    ('py:class', 'Model'),
    ('py:class', 'Database'),
    ('py:class', 'tomlkit.toml_document.TOMLDocument'),
    ('py:class', 'TOMLDocument'),
    ('py:class', 'tomlkit.items.String'),
    ('py:class', 'tomlkit.items.Item'),
    ('py:class', 'playhouse.signals.Model'),
    ('py:class', 'collections.UserDict'),
    ('py:class', "dict[slice(<class 'str'>, <class 'peewee.ModelBase'>, None)]"),
    ('py:class', 'collections.abc.Callable'),
    ('py:class', 'collections.abc.Sequence'),
    ('py:class', 'collections.abc.Iterable'),
    ('py:class', 'collections.abc.MutableMapping'),
    ('py:class', 'collections.abc.Mapping'),
    ('py:class', 're.Pattern'),
    ('py:class', 'pd.DataFrame'),
    ('py:class', 'pandas.DataFrame'),
    ('py:class', 'pandas.core.frame.DataFrame'),
    ('py:class', 'typing._ProtocolMeta'),
    ('py:class', 'sns.FacetGrid'),
    ('py:class', 'seaborn.axisgrid.FacetGrid'),
    ('py:class', 'datetime.date'),
    ('py:class', 'datetime.datetime'),
    ('py:class', 'datetime.timedelta'),
    ('py:class', 'pandas._libs.tslibs.timestamps.Timestamp'),
    ('py:class', 'pandas._libs.tslibs.timedeltas.Timedelta'),
    ('py:class', 'matplotlib.colors.Colormap'),
    ('py:class', 'Colormap'),
    ('py:class', 'click.core.Command'),
    ('py:class', 'click.core.Context'),
    ('py:class', 'click.core.Group'),
]

# find a better solution for the graphviz_dot path. this works for my local
# installation but for sure it won't work for the gitlab container.
# can we set an enviromental variable?
graphviz_dot = r'C:\Program Files\Graphviz\bin\fdp.exe'
graphviz_output_format = 'svg'

templates_path = ['_templates']
exclude_patterns = []

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'sphinx_rtd_theme'
html_static_path = ['_static']
todo_include_todos = True
html_sidebars = {'**': ['localtoc.html', 'sourcelink.html', 'searchbox.html', 'relations.html']}
html_css_files = [
    'css/table_custom.css',
]

autosummary_generate = True
autodoc_default_options = {
    'members': True,
    'member-order': 'groupwise',
    'show-inheritance': True,
    'private-members': True,
}

autodoc_mock_imports = ['pluggy', 'pluggy._manager', 'typing']
autoclass_content = 'both'
autosectionlabel_prefix_document = True

copybutton_exclude = '.linenos, .gp'

match_simple_bash = r'\$ '
match_env_bash = r'\([\w\-]+\)\s*\$ '
match_simple_dos = r'\w\:\\[\w\s\-]*\\*\> '
match_env_dos = r'\([\w\-]*\)\s*\w\:\\[\w\-]*\\*\> '

copybutton_prompt_text = '|'.join([match_simple_bash, match_simple_dos, match_env_dos, match_env_bash])
copybutton_prompt_is_regexp = True


rst_prolog = """
    .. role:: python(code)
        :language: python
"""

external_links = {
    'Google': 'https://google.com',  # matches ":link:`google`", ":link:`Google`", etc
    'pandas': 'https://pandas.pydata.org/',
    'peewee': 'http://docs.peewee-orm.com/',
    'dbeaver': 'https://dbeaver.io/',
    'SQLite': 'https://www.sqlite.org/',
    'ORACLE': 'https://www.oracle.com/',
    'PostgreSQL': 'https://www.postgresql.org/',
    'MySQL': 'https://www.mysql.com/',
    'matplotlib': 'https://matplotlib.org/',
    'seaborn': 'https://seaborn.pydata.org/',
    'numpy': 'https://numpy.org/',
    'toml': 'https://toml.io/',
    'textual': 'https://textual.textualize.io/',
    'jupyter': 'https://jupyter.org/',
    'marimo': 'https://marimo.io/',
    'marlin': 'https://ilcsoft.desy.de/portal/software_packages/marlin/',
    'EUTelescope': 'https://iopscience.iop.org/article/10.1088/1748-0221/15/09/P09020',
    'lcio': 'https://ilcsoft.desy.de/portal/software_packages/lcio/',
    'click': 'https://click.palletsprojects.com/en/stable/',
}
external_links_substitutions = {}

extlinks = {
    'issue': ('https://code.europa.eu/kada/mafw/-/issues%s', 'issue %s'),
}

html_favicon = '_static/images/general/mafw-logo.svg'

## options for latex/pdf output
latex_documents = [
    ('index', 'mafw.tex', 'MAFw: Modular Analysis Framework', 'Antonio Bulgheroni et al.', 'manual', True)
    # (startdocname, targetname, title, author, theme, toctree_only)
]


latex_logo = '_static/images/general/mafw-logo.pdf'
latex_elements = {'papersize': 'a4paper', 'fncychap': r'\usepackage[Glenn]{fncychap}'}

modindex_common_prefix = ['mafw.']

master_doc = 'index'



#  Copyright 2025 European Union
#  Author: Bulgheroni Antonio (antonio.bulgheroni@ec.europa.eu)
#  SPDX-License-Identifier: EUPL-1.2
"""
Provides a container to run configurable and modular analytical tasks.
"""

import itertools
import logging
from pathlib import Path
from typing import TYPE_CHECKING, Any

from pluggy import PluginManager

import mafw.db.std_tables
from mafw import mafw_errors
from mafw.db.std_tables import StandardTable
from mafw.enumerators import ProcessorExitStatus
from mafw.plugin_manager import get_plugin_manager
from mafw.processor import Processor, ProcessorList
from mafw.tools import toml_tools
from mafw.ui.abstract_user_interface import UserInterfaceBase

log = logging.getLogger(__name__)


class MAFwApplication:
    """
    The MAFw Application.

    This class takes care of reading a steering file and from the information retrieved construct a
    :class:`~mafw.processor.ProcessorList` from the processor listed there and execute it.

    It is very practical because any combination of processors can be run without having to write dedicated scripts
    but simply modifying a steering file.

    The application will search for processors not only among the ones available in the MAFw library, but also in all
    other packages exposing processors via the :ref:`plugin mechanism <plugins>`.

    All parameters in the constructor are optional.

    An instance can be created also without the `steering_file`, but such an instance cannot be executed. The steering
    file can be provided in a later stage via the :meth:`init` method or directly to the :meth:`run` method.

    The user interface can be either provided directly in the constructor, or it will be taken from the steering
    file. In the worst case, the fallback :class:`~mafw.ui.console_user_interface.ConsoleInterface` will be used.

    The plugin manager, if not provided, the global plugin manager will be retrieved from the
    :func:`~mafw.plugin_manager.get_plugin_manager`.

    A simple example is provided here below:

    .. code-block:: python
        :name: MAFwApplication_run
        :caption: Creation and execution of a MAFwApplication

        import logging
        from pathlib import Path

        from mafw.runner import MAFwApplication

        log = logging.getLogger(__name__)

        # put here your steering file
        steering_file = Path('path_to_my_steering_file.toml')

        try:
            # create the app
            app = MAFwApplication(steering_file)

            # run it!
            app.run()

        except Exception as e:
            log.error('An error occurred!')
            log.exception(e)
    """

    __default_ui__: str = 'rich'

    def __init__(
        self,
        steering_file: Path | str | None = None,
        user_interface: UserInterfaceBase | type[UserInterfaceBase] | str | None = None,
        plugin_manager: PluginManager | None = None,
    ):
        """
        Constructor parameters:

        :param steering_file: The path to the steering file.
        :type steering_file: Path | str, Optional
        :param user_interface: The user interface to be used by the application.
        :type user_interface: UserInterfaceBase | type[UserInterfaceBase] | str, Optional
        :param plugin_manager: The plugin manager.
        :type plugin_manager: PluginManager, Optional
        """
        #: the name of the application instance
        self.name = self.__class__.__name__
        self._configuration_dict: dict[str, Any] = {}

        #: the plugin manager of the application instance
        self.plugin_manager = plugin_manager or get_plugin_manager()

        #: the exit status of the application
        self.exit_status = ProcessorExitStatus.Successful

        self.user_interface: UserInterfaceBase | None
        if user_interface is not None:
            if isinstance(user_interface, UserInterfaceBase):
                self.user_interface = user_interface
            else:  # if isinstance(user_interface, str):
                if TYPE_CHECKING:
                    assert isinstance(user_interface, str)
                self.user_interface = self.get_user_interface(user_interface)
        else:
            self.user_interface = None

        if steering_file is None:
            self._initialized = False
            self.steering_file = None
        else:
            self.steering_file = steering_file if isinstance(steering_file, Path) else Path(steering_file)
            self.init(self.steering_file)

    def get_user_interface(self, user_interface: str) -> UserInterfaceBase:
        """
        Retrieves the user interface from the plugin managers.

        User interfaces are exposed via the plugin manager.

        If the requested `user_interface` is not available, then the fallback console interface is used.

        :param user_interface: The name of the user interface to be used. Normally rich or console.
        :type user_interface: str
        """
        available_ui_list = list(itertools.chain(*self.plugin_manager.hook.register_user_interfaces()))
        available_ui: dict[str, type[UserInterfaceBase]] = {p.name: p for p in available_ui_list}
        if user_interface in available_ui:
            ui_type = available_ui[user_interface]
        else:
            log.warning('User interface %s is not available. Using console.' % user_interface)
            ui_type = available_ui['console']
        return ui_type()

    def run(self, steering_file: Path | str | None = None) -> ProcessorExitStatus:
        """
        Runs the application.

        This method builds the :class:`~mafw.processor.ProcessorList` with the processors listed in the steering file
        and launches its execution.

        A steering file can be provided at this stage if it was not done before.

        :param steering_file: The steering file. Defaults to None.
        :type steering_file: Path | str, Optional
        :raises RunnerNotInitialized: if the application has not been initialized. Very likely a steering file was
            never provided.
        :raises UnknownProcessor: if a processor listed in the steering file is not available in the plugin library.
        """
        if steering_file is None and not self._initialized:
            log.error('%s is not initialized. Have you provided a steering file?' % self.name)
            raise mafw_errors.RunnerNotInitialized()

        if steering_file is not None and steering_file != self.steering_file:
            self.init(steering_file)

        description = self._configuration_dict.get('analysis_description', None)
        processors_to_run = self._configuration_dict.get('processors_to_run', [])

        available_processors_list = list(itertools.chain(*self.plugin_manager.hook.register_processors()))
        available_processors: dict[str, type[Processor]] = {p.__name__: p for p in available_processors_list}

        external_standard_table_list = list(itertools.chain(*self.plugin_manager.hook.register_standard_tables()))
        external_standard_table: dict[str, type[StandardTable]] = {
            t.__qualname__: t for t in external_standard_table_list
        }
        mafw.db.std_tables.standard_tables.update(external_standard_table)

        for processor in processors_to_run:
            if processor not in available_processors:
                log.critical(
                    'Processor %s is not available. Check your plugin configuration or your steering file.' % processor
                )
                raise mafw_errors.UnknownProcessor(processor)

        processor_list = ProcessorList(
            name=self.name,
            description=description,
            user_interface=self.user_interface,
            database_conf=self._configuration_dict.get('DBConfiguration', None),
        )
        for processor in processors_to_run:
            processor_list.append(available_processors[processor](config=self._configuration_dict))

        return processor_list.execute()

    def init(self, steering_file: Path | str) -> None:
        """
        Initializes the application.

        This method is normally automatically invoked by the class constructor.
        It can be called in a later moment to force the parsing of the provided steering file.

        :param steering_file: The path to the steering file.
        :type steering_file: Path | str
        """
        if isinstance(steering_file, str):
            steering_file = Path(steering_file)

        self.steering_file = steering_file
        self._configuration_dict = toml_tools.load_steering_file(steering_file)

        if self.user_interface is None:
            self.user_interface = self.get_user_interface(self._configuration_dict['UserInterface']['interface'])

        self._initialized = True
        self.name = self._configuration_dict.get('analysis_name', 'mafw analysis')



#  Copyright 2025 European Union
#  Author: Bulgheroni Antonio (antonio.bulgheroni@ec.europa.eu)
#  SPDX-License-Identifier: EUPL-1.2
"""
Unit tests for the db_model module.
"""

from unittest.mock import Mock, patch

import pytest
from peewee import AutoField, DatabaseProxy, ModelInsert, TextField

from mafw.db.db_model import MAFwBaseModel, MAFwBaseModelDoesNotExist, database_proxy
from mafw.db.trigger import Trigger
from mafw.mafw_errors import UnsupportedDatabaseError


class TestMAFwBaseModelDoesNotExist:
    """Test cases for MAFwBaseModelDoesNotExist exception."""

    def test_exception_inheritance(self):
        """Test that the exception inherits from MAFwException."""
        from mafw.mafw_errors import MAFwException

        assert issubclass(MAFwBaseModelDoesNotExist, MAFwException)

    def test_exception_can_be_raised(self):
        """Test that the exception can be raised and caught."""
        with pytest.raises(MAFwBaseModelDoesNotExist):
            raise MAFwBaseModelDoesNotExist('Test message')


class TestDatabaseProxy:
    """Test cases for database_proxy."""

    def test_database_proxy_is_instance(self):
        """Test that database_proxy is an instance of DatabaseProxy."""
        assert isinstance(database_proxy, DatabaseProxy)


class TestMAFwBaseModel:
    """Test cases for MAFwBaseModel class."""

    @pytest.fixture
    def mock_database(self):
        """Create a mock database for testing."""
        db = Mock()
        db.execute_sql = Mock()
        db.drop_tables = Mock()
        return db

    @pytest.fixture
    def mock_trigger(self):
        """Create a mock trigger for testing."""
        trigger = Mock(spec=Trigger)
        trigger.trigger_name = 'test_trigger'
        trigger.set_database = Mock()
        trigger.create = Mock(return_value='CREATE TRIGGER test_trigger...')
        trigger.drop = Mock(return_value='DROP TRIGGER test_trigger')
        return trigger

    @pytest.fixture
    def sample_model_class(self, mock_database):
        """Create a sample model class for testing."""

        class SampleModel(MAFwBaseModel):
            id = AutoField(primary_key=True)
            name = TextField()

            class Meta:
                database = mock_database
                legacy_table_names = False

        # Mock the _meta attribute
        SampleModel._meta = Mock()
        SampleModel._meta.database = mock_database
        SampleModel._meta.composite_key = False
        SampleModel._meta.primary_key = Mock()
        SampleModel._meta.primary_key.name = 'id'
        SampleModel._meta.fields = {'id': Mock(name='id'), 'name': Mock(name='name')}

        return SampleModel

    def test_triggers_default_implementation(self):
        """Test that the default triggers method returns an empty list."""
        triggers = MAFwBaseModel.triggers()
        assert triggers == []
        assert isinstance(triggers, list)

    def test_triggers_can_be_overridden(self, mock_trigger):
        """Test that triggers method can be overridden in subclasses."""

        class CustomModel(MAFwBaseModel):
            @classmethod
            def triggers(cls):
                return [mock_trigger]

        triggers = CustomModel.triggers()
        assert len(triggers) == 1
        assert triggers[0] == mock_trigger

    @patch('mafw.db.db_model.Model.create_table')
    def test_create_table_without_triggers(self, mock_super_create, sample_model_class):
        """Test create_table when no triggers are defined."""
        sample_model_class.triggers = Mock(return_value=[])

        sample_model_class.create_table(safe=True, test_option='value')

        mock_super_create.assert_called_once_with(True, test_option='value')
        sample_model_class._meta.database.execute_sql.assert_not_called()

    @patch('mafw.db.db_model.Model.create_table')
    def test_create_table_with_triggers_success(self, mock_super_create, sample_model_class, mock_trigger):
        """Test successful create_table with triggers."""
        sample_model_class.triggers = Mock(return_value=[mock_trigger])

        sample_model_class.create_table()

        mock_super_create.assert_called_once_with(True)
        mock_trigger.set_database.assert_called_once_with(sample_model_class._meta.database)
        mock_trigger.create.assert_called_once()
        sample_model_class._meta.database.execute_sql.assert_called_once_with(mock_trigger.create.return_value)

    @patch('mafw.db.db_model.Model.create_table')
    def test_create_table_with_unsupported_trigger(self, mock_super_create, sample_model_class, mock_trigger):
        """Test create_table with unsupported trigger (should warn and continue)."""
        sample_model_class.triggers = Mock(return_value=[mock_trigger])
        sample_model_class._meta.database.execute_sql.side_effect = UnsupportedDatabaseError('Unsupported trigger')

        with pytest.warns(UserWarning, match='Skipping unsupported trigger'):
            sample_model_class.create_table()

        mock_super_create.assert_called_once()
        sample_model_class._meta.database.drop_tables.assert_not_called()

    @patch('mafw.db.db_model.Model.create_table')
    def test_create_table_with_trigger_failure_cleanup(self, mock_super_create, sample_model_class, mock_trigger):
        """Test create_table cleanup when trigger creation fails."""
        sample_model_class.triggers = Mock(return_value=[mock_trigger])
        sample_model_class._meta.database.execute_sql.side_effect = Exception('Trigger creation failed')

        with pytest.raises(Exception, match='Trigger creation failed'):
            sample_model_class.create_table()

        # Verify cleanup was attempted
        sample_model_class._meta.database.drop_tables.assert_called_once_with([sample_model_class], safe=True)
        mock_trigger.drop.assert_called_once_with(True)
        sample_model_class._meta.database.execute_sql.assert_called()

    @patch('mafw.db.db_model.Model.create_table')
    def test_create_table_cleanup_ignores_drop_errors(self, mock_super_create, sample_model_class, mock_trigger):
        """Test that cleanup ignores errors when dropping triggers."""
        sample_model_class.triggers = Mock(return_value=[mock_trigger])
        sample_model_class._meta.database.execute_sql.side_effect = Exception('Trigger creation failed')

        # Make trigger.drop raise an exception - should be ignored
        mock_trigger.drop.side_effect = Exception('Drop failed')

        with pytest.raises(Exception, match='Trigger creation failed'):
            sample_model_class.create_table()

        # Should still attempt cleanup despite drop failure
        sample_model_class._meta.database.drop_tables.assert_called_once()
        mock_trigger.drop.assert_called_once()

    @patch('mafw.db.db_model.Model.create_table')
    def test_create_table_multiple_triggers(self, mock_super_create, sample_model_class):
        """Test create_table with multiple triggers."""
        trigger1 = Mock(spec=Trigger)
        trigger1.trigger_name = 'trigger1'
        trigger1.set_database = Mock()
        trigger1.create = Mock(return_value='CREATE TRIGGER trigger1...')
        trigger1.drop = Mock(return_value='DROP TRIGGER trigger1')

        trigger2 = Mock(spec=Trigger)
        trigger2.trigger_name = 'trigger2'
        trigger2.set_database = Mock()
        trigger2.create = Mock(return_value='CREATE TRIGGER trigger2...')
        trigger2.drop = Mock(return_value='DROP TRIGGER trigger2')

        sample_model_class.triggers = Mock(return_value=[trigger1, trigger2])

        sample_model_class.create_table()

        # Verify both triggers were processed
        trigger1.set_database.assert_called_once_with(sample_model_class._meta.database)
        trigger2.set_database.assert_called_once_with(sample_model_class._meta.database)
        assert sample_model_class._meta.database.execute_sql.call_count == 2

    @pytest.fixture
    def mock_model_insert(self):
        """Create a mock ModelInsert for testing."""
        mock_insert = Mock(spec=ModelInsert)
        mock_on_conflict = Mock(spec=ModelInsert)
        mock_insert.on_conflict = Mock(return_value=mock_on_conflict)
        return mock_insert, mock_on_conflict

    @patch('mafw.db.db_model.cast')
    def test_std_upsert_simple_primary_key(self, mock_cast, sample_model_class, mock_model_insert):
        """Test std_upsert with simple primary key."""
        mock_insert, mock_on_conflict = mock_model_insert
        mock_cast.side_effect = [sample_model_class, mock_on_conflict]

        with patch.object(sample_model_class, 'insert', return_value=mock_insert):
            result = sample_model_class.std_upsert({'name': 'test'}, id=1)

            sample_model_class.insert.assert_called_once_with({'name': 'test'}, id=1)
            mock_insert.on_conflict.assert_called_once()

            # Verify the on_conflict call arguments
            call_args = mock_insert.on_conflict.call_args
            assert 'conflict_target' in call_args.kwargs
            assert 'preserve' in call_args.kwargs

            assert result == mock_on_conflict

    @patch('mafw.db.db_model.cast')
    def test_std_upsert_composite_key(self, mock_cast, sample_model_class, mock_model_insert):
        """Test std_upsert with composite primary key."""
        mock_insert, mock_on_conflict = mock_model_insert
        mock_cast.side_effect = [sample_model_class, mock_on_conflict]

        # Setup composite key scenario
        sample_model_class._meta.composite_key = True
        sample_model_class._meta.primary_key = Mock()
        sample_model_class._meta.primary_key.field_names = ['id', 'type']
        sample_model_class._meta.fields = {'id': Mock(name='id'), 'type': Mock(name='type'), 'name': Mock(name='name')}

        with patch.object(sample_model_class, 'insert', return_value=mock_insert):
            result = sample_model_class.std_upsert(name='test')

            sample_model_class.insert.assert_called_once_with(None, name='test')
            mock_insert.on_conflict.assert_called_once()
            assert result == mock_on_conflict

    @patch('mafw.db.db_model.cast')
    def test_std_upsert_many_simple_key(self, mock_cast, sample_model_class, mock_model_insert):
        """Test std_upsert_many with simple primary key."""
        mock_insert, mock_on_conflict = mock_model_insert
        mock_cast.side_effect = [sample_model_class, mock_on_conflict]

        rows = [{'id': 1, 'name': 'test1'}, {'id': 2, 'name': 'test2'}]

        with patch.object(sample_model_class, 'insert_many', return_value=mock_insert):
            result = sample_model_class.std_upsert_many(rows, ['id', 'name'])

            sample_model_class.insert_many.assert_called_once_with(rows, ['id', 'name'])
            mock_insert.on_conflict.assert_called_once()
            assert result == mock_on_conflict

    @patch('mafw.db.db_model.cast')
    def test_std_upsert_many_composite_key(self, mock_cast, sample_model_class, mock_model_insert):
        """Test std_upsert_many with composite primary key."""
        mock_insert, mock_on_conflict = mock_model_insert
        mock_cast.side_effect = [sample_model_class, mock_on_conflict]

        # Setup composite key scenario
        sample_model_class._meta.composite_key = True
        sample_model_class._meta.primary_key = Mock()
        sample_model_class._meta.primary_key.field_names = ['id', 'type']
        sample_model_class._meta.fields = {'id': Mock(name='id'), 'type': Mock(name='type'), 'name': Mock(name='name')}

        rows = [(1, 'A', 'test1'), (2, 'B', 'test2')]

        with patch.object(sample_model_class, 'insert_many', return_value=mock_insert):
            result = sample_model_class.std_upsert_many(rows, ['id', 'type', 'name'])

            sample_model_class.insert_many.assert_called_once_with(rows, ['id', 'type', 'name'])
            mock_insert.on_conflict.assert_called_once()
            assert result == mock_on_conflict

    @pytest.mark.parametrize(
        'safe_param,expected_safe',
        [
            (True, True),
            (False, False),
            (None, True),  # Default value
        ],
    )
    @patch('mafw.db.db_model.Model.create_table')
    def test_create_table_safe_parameter(self, mock_super_create, sample_model_class, safe_param, expected_safe):
        """Test create_table with different safe parameter values."""
        sample_model_class.triggers = Mock(return_value=[])

        if safe_param is None:
            sample_model_class.create_table()
        else:
            sample_model_class.create_table(safe=safe_param)

        mock_super_create.assert_called_once_with(expected_safe)

    def test_meta_class_attributes(self):
        """Test that Meta class has correct default attributes."""
        assert MAFwBaseModel._meta.database == database_proxy
        assert MAFwBaseModel._meta.legacy_table_names is False

    def test_model_inheritance(self):
        """Test that MAFwBaseModel properly inherits from peewee Model."""
        # Import here to avoid circular imports in the actual test
        from playhouse.signals import Model

        assert issubclass(MAFwBaseModel, Model)

    @patch('mafw.db.db_model.warnings.warn')
    @patch('mafw.db.db_model.Model.create_table')
    def test_unsupported_database_warning_message(self, mock_super_create, mock_warn, sample_model_class, mock_trigger):
        """Test that the correct warning message is generated for unsupported database errors."""
        sample_model_class.triggers = Mock(return_value=[mock_trigger])
        error_msg = "This database doesn't support triggers"
        sample_model_class._meta.database.execute_sql.side_effect = UnsupportedDatabaseError(error_msg)

        sample_model_class.create_table()

        mock_warn.assert_called_once_with(f'Skipping unsupported trigger {mock_trigger.trigger_name}: {error_msg}')

    @patch('mafw.db.db_model.cast')
    def test_std_upsert_without_data_parameter(self, mock_cast, sample_model_class, mock_model_insert):
        """Test std_upsert when __data parameter is None."""
        mock_insert, mock_on_conflict = mock_model_insert
        mock_cast.side_effect = [sample_model_class, mock_on_conflict]

        with patch.object(sample_model_class, 'insert', return_value=mock_insert):
            result = sample_model_class.std_upsert(None, name='test')

            sample_model_class.insert.assert_called_once_with(None, name='test')
            assert result == mock_on_conflict

    @patch('mafw.db.db_model.cast')
    def test_std_upsert_many_without_fields_parameter(self, mock_cast, sample_model_class, mock_model_insert):
        """Test std_upsert_many when fields parameter is None."""
        mock_insert, mock_on_conflict = mock_model_insert
        mock_cast.side_effect = [sample_model_class, mock_on_conflict]
        rows = [{'id': 1, 'name': 'test'}]

        with patch.object(sample_model_class, 'insert_many', return_value=mock_insert):
            result = sample_model_class.std_upsert_many(rows, None)

            sample_model_class.insert_many.assert_called_once_with(rows, None)
            assert result == mock_on_conflict


class TestModuleConstants:
    """Test module-level constants and imports."""

    def test_database_proxy_exists(self):
        """Test that database_proxy is properly defined."""
        from mafw.db.db_model import database_proxy

        assert database_proxy is not None
        assert isinstance(database_proxy, DatabaseProxy)


@pytest.mark.integration_test
class TestIntegration:
    """Integration tests that test multiple components together."""

    @pytest.fixture
    def complete_model_setup(self):
        """Setup a complete model with all components for integration testing."""
        db = Mock()
        db.execute_sql = Mock()
        db.drop_tables = Mock()

        trigger = Mock(spec=Trigger)
        trigger.trigger_name = 'integration_trigger'
        trigger.set_database = Mock()
        trigger.create = Mock(return_value='CREATE TRIGGER integration_trigger...')
        trigger.drop = Mock(return_value='DROP TRIGGER integration_trigger')

        class IntegrationModel(MAFwBaseModel):
            id = AutoField(primary_key=True)
            name = TextField()

            @classmethod
            def triggers(cls):
                return [trigger]

            class Meta:
                database = db

        # Setup _meta properly
        IntegrationModel._meta = Mock()
        IntegrationModel._meta.database = db
        IntegrationModel._meta.composite_key = False
        IntegrationModel._meta.primary_key = Mock(name='id')
        IntegrationModel._meta.fields = {'id': Mock(name='id'), 'name': Mock(name='name')}

        return IntegrationModel, db, trigger

    @pytest.fixture
    def mock_model_insert(self):
        """Create a mock ModelInsert for testing."""
        mock_insert = Mock(spec=ModelInsert)
        mock_on_conflict = Mock(spec=ModelInsert)
        mock_insert.on_conflict = Mock(return_value=mock_on_conflict)
        return mock_insert, mock_on_conflict

    @patch('mafw.db.db_model.Model.create_table')
    @patch('mafw.db.db_model.cast')
    def test_complete_workflow_success(self, mock_cast, mock_super_create, complete_model_setup, mock_model_insert):
        """Test the complete workflow from table creation to upsert operations."""
        model_class, db, trigger = complete_model_setup
        mock_insert, mock_on_conflict = mock_model_insert
        mock_cast.side_effect = [model_class, model_class, mock_on_conflict]
        # Test table creation
        model_class.create_table()

        # Verify table and trigger creation
        mock_super_create.assert_called_once()
        trigger.set_database.assert_called_once_with(db)
        db.execute_sql.assert_called_once()

        with patch.object(model_class, 'insert', return_value=mock_insert):
            result = model_class.std_upsert(name='integration_test')

            assert result == mock_on_conflict
            model_class.insert.assert_called_once()
            mock_insert.on_conflict.assert_called_once()



#  Copyright 2025 European Union
#  Author: Bulgheroni Antonio (antonio.bulgheroni@ec.europa.eu)
#  SPDX-License-Identifier: EUPL-1.2
"""
Module implements active variable for classes.
"""

from __future__ import annotations

from typing import Generic, Type, TypeVar, cast

ActiveType = TypeVar('ActiveType')
"""A type for templating the Active class."""

ActivableType = TypeVar('ActivableType')
"""A type for  all classes that can include an Active."""


class Active(Generic[ActiveType]):
    """
    A descriptor class to make class variable **active**.

    When assigned to a class variable, any change of this value will trigger a call back to a specific function.

    Here is a clarifying example.

    .. code-block:: python

        class Person:
            age = Active()

            def __init__(self, age):
                self.age = age

            def on_age_change(self, old_value, new_value):
                # callback invoked every time the value of age is changed
                # do something with the old and the new age
                pass

            def on_age_set(self, value):
                # callback invoked every time the value on age is set to the same
                # value as before.
                pass

            def on_age_get(self, value):
                # callback invoked every time the value of age is asked.
                # not really useful, but...
                pass


    Once you have assigned an Active to a class member, you need to implement the callback in your class.
    If you do not implement them, the code will run without problems.

    The three callbacks have the signature described in the example.

        * on_[var_name]_change(self, old, new)
        * on_[var_name]_set(self, value)
        * on_[var_name]_get(self, value)
    """

    def __init__(self, default: ActiveType | None = None) -> None:
        """
        Constructor parameter:

        :param default: Initial value of the Active value. Defaults to None.
        :type default: ActiveType
        """
        self.default = default

        # the name of the three callbacks.
        # they will be assigned after we know the name of the variable.
        self._change_call_back_name: str
        self._set_callback_name: str
        self._get_callback_name: str

    def __set_name__(self, obj: Type[ActivableType], name: str) -> None:
        # this is the public name of the class variable
        self.public_name = name
        # this is the name where we will be storing the value in the owner class
        self.private_name = 'active_' + name

        # check if the owner does not have an attribute named after private name
        # if so create it and assign it the default value
        if not hasattr(obj, self.private_name):
            setattr(obj, self.private_name, self.default)

        # now you can prepare the name for the callbacks
        self._init_callbacks()

    def _init_callbacks(self) -> None:
        # now we know the name of the variable, so we can create the callback names as well
        self._change_callback_name = f'on_{self.public_name}_change'
        self._set_callback_name = f'on_{self.public_name}_set'
        self._get_callback_name = f'on_{self.public_name}_get'

    def __get__(self, obj: ActivableType, obj_type: Type[ActivableType]) -> Active[ActiveType] | ActiveType:
        if obj is None:
            # obj is None means that we are invoking the descriptor via the class and not the instance
            # in this case we cannot return the private value and we return the descriptor itself.
            return self

        # if the object does not have an attribute named with our private_name, it means this is the first use.
        # we need to initialize the descriptor
        if not hasattr(obj, self.private_name):
            setattr(obj, self.private_name, self.default)

        # the value is stored in the private_name attribute.
        value = getattr(obj, self.private_name)

        # check if there is a get callback and if yes call it
        if hasattr(obj, self._get_callback_name):
            getattr(obj, self._get_callback_name)(value)

        return cast(ActiveType, value)

    def __set__(self, obj: ActivableType, value: ActiveType) -> None:
        # this is the current value
        current_value = getattr(obj, self.private_name)
        # now set the new value
        setattr(obj, self.private_name, value)

        if current_value != value:
            # the value really changed, call the change callback if it exists:
            if hasattr(obj, self._change_callback_name):
                getattr(obj, self._change_callback_name)(current_value, value)
        else:
            # the value did not change, but it was set anyhow. call the set callback if it exists
            if hasattr(obj, self._set_callback_name):
                getattr(obj, self._set_callback_name)(value)



#  Copyright 2025 European Union
#  Author: Bulgheroni Antonio (antonio.bulgheroni@ec.europa.eu)
#  SPDX-License-Identifier: EUPL-1.2
import sys
import warnings
from typing import List
from unittest.mock import patch

import pytest

from mafw.mafw_errors import MissingOptionalDependency


class TestFailedSNSPlotterImports:
    """Test class for testing import failure scenarios in sns_plotter module."""

    def setup_method(self):
        """Setup method to store original state before each test."""
        # Store the original import function
        if isinstance(__builtins__, dict):
            self.original_import = __builtins__['__import__']
        else:
            self.original_import = __builtins__.__import__

        # Store references to all seaborn-related modules and the sns_plotter module
        self.original_modules = {}

        # Store ALL matplotlib / seaborn-related modules that might be affected
        seaborn_modules = [k for k in sys.modules.keys() if k.startswith('seaborn')]
        pandas_modules = [k for k in sys.modules.keys() if k.startswith('pandas')]
        matplotlib_modules = [k for k in sys.modules.keys() if k.startswith('matplotlib')]
        sns_plotter_modules = ['mafw.processor_library.sns_plotter', 'sns_plotter']

        all_modules_to_store = matplotlib_modules + pandas_modules + seaborn_modules + sns_plotter_modules

        for module_name in all_modules_to_store:
            self.original_modules[module_name] = sys.modules.get(module_name)

        # Also store the specific sns reference from sns_plotter if it exists
        if 'mafw.processor_library.sns_plotter' in sys.modules:
            sns_plotter_module = sys.modules['mafw.processor_library.sns_plotter']
            if hasattr(sns_plotter_module, 'sns'):
                self.original_sns_reference = sns_plotter_module.sns
            else:
                self.original_sns_reference = None
        else:
            self.original_sns_reference = None

    def teardown_method(self):
        """Teardown method to restore original state after each test."""

        # Restore original __import__ function first
        if isinstance(__builtins__, dict):
            __builtins__['__import__'] = self.original_import
        else:
            __builtins__.__import__ = self.original_import

        # Remove all modules that we're going to restore to ensure clean state
        for module_name in self.original_modules.keys():
            if module_name in sys.modules:
                del sys.modules[module_name]

        # Restore all original modules in the correct order
        # First restore matplotlib / seaborn modules
        # seaborn_modules = {k: v for k, v in self.original_modules.items() if k.startswith('seaborn')}
        # for module_name, module in seaborn_modules.items():
        for module_name, module in self.original_modules.items():
            if module is not None:
                sys.modules[module_name] = module

        # Then restore sns_plotter modules
        sns_plotter_modules = {k: v for k, v in self.original_modules.items() if 'sns_plotter' in k}
        for module_name, module in sns_plotter_modules.items():
            if module is not None:
                sys.modules[module_name] = module

        # Restore the sns reference in sns_plotter module if needed
        if self.original_sns_reference is not None and 'mafw.processor_library.sns_plotter' in sys.modules:
            sns_plotter_module = sys.modules['mafw.processor_library.sns_plotter']
            if hasattr(sns_plotter_module, 'sns'):
                sns_plotter_module.sns = self.original_sns_reference

        # Clear stored modules
        self.original_modules.clear()

    def _remove_modules(self, modules_to_remove: List[str]) -> None:
        """
        Helper method to remove modules from sys.modules.
        Note: We don't store them here since they're already stored in setup_method.

        Args:
            modules_to_remove: List of module names to remove
        """
        for module_name in modules_to_remove:
            if module_name in sys.modules:
                del sys.modules[module_name]

    @pytest.mark.parametrize(
        'missing_modules,expected_in_message',
        [
            # Test individual missing modules
            (['matplotlib'], 'seaborn'),
            (['matplotlib.pyplot'], 'seaborn'),
            (['matplotlib.colors'], 'seaborn'),
            (['pandas'], 'seaborn'),
            (['seaborn'], 'seaborn'),
            # Test combinations of missing modules
            (['matplotlib', 'pandas'], 'seaborn'),
            (['matplotlib', 'seaborn'], 'seaborn'),
            (['pandas', 'seaborn'], 'seaborn'),
            (['matplotlib', 'pandas', 'seaborn'], 'seaborn'),
            # Test matplotlib submodules
            (['matplotlib.pyplot', 'matplotlib.colors'], 'seaborn'),
            (['matplotlib.typing'], 'seaborn'),
        ],
    )
    def test_missing_dependencies_raise_warning_and_exception(
        self, missing_modules: List[str], expected_in_message: str
    ):
        """
        Test that missing dependencies raise appropriate warnings and exceptions.

        Args:
            missing_modules: List of module names to simulate as missing
            expected_in_message: String expected to be in the warning message
        """
        # Remove sns_plotter modules to force reimport
        self._remove_modules(['mafw.processor_library.sns_plotter', 'sns_plotter'])

        # Remove the specified dependency modules
        self._remove_modules(missing_modules)

        # Mock the import mechanism to raise ImportError for missing modules
        def mock_import(name, *args, **kwargs):
            if any(name.startswith(missing_module) for missing_module in missing_modules):
                raise ImportError(f"**MOCK** No module named '{name}'")
            return self.original_import(name, *args, **kwargs)

        # Use context manager to ensure proper cleanup of the mock
        with patch.dict(__builtins__, {'__import__': mock_import}):
            # Test that importing sns_plotter raises ImportError and warning
            with pytest.warns(MissingOptionalDependency) as warning_info:
                with pytest.raises(ImportError):
                    import mafw.processor_library.sns_plotter  # noqa: F401

        # Verify warning message content
        warning_message = str(warning_info[0].message)
        assert expected_in_message in warning_message
        assert 'pip install mafw[seaborn]' in warning_message
        assert 'required dependencies' in warning_message

    def test_successful_import_when_all_dependencies_available(self):
        """Test that import succeeds when all dependencies are available."""
        # This test ensures our test setup doesn't break normal imports
        # First, make sure we can import the required modules
        try:
            import matplotlib
            import matplotlib.colors
            import matplotlib.pyplot  # noqa: F401
            import pandas  # noqa: F401
            import seaborn  # noqa: F401
        except ImportError:
            pytest.skip('Required dependencies not available in test environment')

        # Now test that sns_plotter can be imported without warnings
        with warnings.catch_warnings():
            warnings.simplefilter('error')  # Turn warnings into errors
            try:
                import mafw.processor_library.sns_plotter  # noqa: F401
                # If we get here, no warnings were raised
            except Warning as w:
                if isinstance(w, MissingOptionalDependency):
                    pytest.fail('MissingOptionalDependency warning raised when all dependencies available')
                else:
                    # Re-raise other warnings as they might be expected
                    raise
            except ImportError:
                # This might happen if there are other import issues in the module
                # that aren't related to the dependencies we're testing
                pass


class TestFailedPandasToolsImports:
    """Test class for testing import failure scenarios in sns_plotter module."""

    def setup_method(self):
        """Setup method to store original state before each test."""
        # Store the original import function
        if isinstance(__builtins__, dict):
            self.original_import = __builtins__['__import__']
        else:
            self.original_import = __builtins__.__import__

        # Store references to all seaborn-related modules and the sns_plotter module
        self.original_modules = {}

        # Store ALL pandas that might be affected
        pandas_modules = [k for k in sys.modules.keys() if k.startswith('pandas')]

        all_modules_to_store = pandas_modules

        for module_name in all_modules_to_store:
            self.original_modules[module_name] = sys.modules.get(module_name)

        # Also store the specific sns reference from sns_plotter if it exists
        if 'mafw.tools.pandas_tools' in sys.modules:
            pandas_tool_module = sys.modules['mafw.tools.pandas_tools']
            if hasattr(pandas_tool_module, 'pd'):
                self.original_pandas_reference = pandas_tool_module.pd
            else:
                self.original_pandas_reference = None
        else:
            self.original_pandas_reference = None

    def teardown_method(self):
        """Teardown method to restore original state after each test."""

        # Restore original __import__ function first
        if isinstance(__builtins__, dict):
            __builtins__['__import__'] = self.original_import
        else:
            __builtins__.__import__ = self.original_import

        # Remove all modules that we're going to restore to ensure clean state
        for module_name in self.original_modules.keys():
            if module_name in sys.modules:
                del sys.modules[module_name]

        # Restore all original modules in the correct order
        for module_name, module in self.original_modules.items():
            if module is not None:
                sys.modules[module_name] = module

        # Then restore sns_plotter modules
        pandas_tool_module = {k: v for k, v in self.original_modules.items() if 'pandas_tool' in k}
        for module_name, module in pandas_tool_module.items():
            if module is not None:
                sys.modules[module_name] = module

        # Restore the sns reference in sns_plotter module if needed
        if self.original_pandas_reference is not None and 'mafw.tools.pandas_tools' in sys.modules:
            pandas_tool_module = sys.modules['mafw.tools.pandas_tools']
            if hasattr(pandas_tool_module, 'pd'):
                pandas_tool_module.pd = self.original_pandas_reference

        # Clear stored modules
        self.original_modules.clear()

    def _remove_modules(self, modules_to_remove: List[str]) -> None:
        """
        Helper method to remove modules from sys.modules.
        Note: We don't store them here since they're already stored in setup_method.

        Args:
            modules_to_remove: List of module names to remove
        """
        for module_name in modules_to_remove:
            if module_name in sys.modules:
                del sys.modules[module_name]

    @pytest.mark.parametrize(
        'missing_modules,expected_in_message',
        [
            # Test individual missing modules
            (['pandas'], 'seaborn'),
        ],
    )
    def test_missing_dependencies_raise_warning_and_exception(
        self, missing_modules: List[str], expected_in_message: str
    ):
        """
        Test that missing dependencies raise appropriate warnings and exceptions.

        Args:
            missing_modules: List of module names to simulate as missing
            expected_in_message: String expected to be in the warning message
        """
        # Remove sns_plotter modules to force reimport
        self._remove_modules(['mafw.tools.pandas_tools', 'pandas_tools'])

        # Remove the specified dependency modules
        self._remove_modules(missing_modules)

        # Mock the import mechanism to raise ImportError for missing modules
        def mock_import(name, *args, **kwargs):
            if any(name.startswith(missing_module) for missing_module in missing_modules):
                raise ImportError(f"**MOCK** No module named '{name}'")
            return self.original_import(name, *args, **kwargs)

        # Use context manager to ensure proper cleanup of the mock
        with patch.dict(__builtins__, {'__import__': mock_import}):
            # Test that importing sns_plotter raises ImportError and warning
            with pytest.warns(MissingOptionalDependency) as warning_info:
                with pytest.raises(ImportError):
                    import mafw.tools.pandas_tools  # noqa: F401

        # Verify warning message content
        warning_message = str(warning_info[0].message)
        assert expected_in_message in warning_message
        assert 'pip install mafw[seaborn]' in warning_message
        assert 'required dependencies' in warning_message

    def test_successful_import_when_all_dependencies_available(self):
        """Test that import succeeds when all dependencies are available."""
        # This test ensures our test setup doesn't break normal imports
        # First, make sure we can import the required modules
        try:
            import pandas  # noqa: F401
        except ImportError:
            pytest.skip('Required dependencies not available in test environment')

        # Now test that sns_plotter can be imported without warnings
        with warnings.catch_warnings():
            warnings.simplefilter('error')  # Turn warnings into errors
            try:
                import mafw.tools.pandas_tools  # noqa: F401
                # If we get here, no warnings were raised
            except Warning as w:
                if isinstance(w, MissingOptionalDependency):
                    pytest.fail('MissingOptionalDependency warning raised when all dependencies available')
                else:
                    # Re-raise other warnings as they might be expected
                    raise
            except ImportError:
                # This might happen if there are other import issues in the module
                # that aren't related to the dependencies we're testing
                pass



#  Copyright 2025 European Union
#  Author: Bulgheroni Antonio (antonio.bulgheroni@ec.europa.eu)
#  SPDX-License-Identifier: EUPL-1.2
"""
Module provides customised model fields specific for MAFw.
"""

from pathlib import Path
from typing import Any

from peewee import FieldAccessor, TextField

import mafw.tools.file_tools


class FileNameFieldAccessor(FieldAccessor):
    """
    A field accessor specialized for filename fields.

    In the constructor of the :class:`FileNameField` and subclasses, the user can specify the name of a checksum
    field linked to this filename. This is very useful because in this way, the user does not have to manually assign
    any value to this field that will simply be automatically updated when the filename field is updated.

    The user can disable this automatic feature either removing the link in the :class:`FileNameField` or simply
    assigning a value to the :class:`FileChecksumField`.
    """

    # noinspection PyProtectedMember
    def __set__(self, instance: Any, value: Any) -> None:
        """
        Sets the value of field in the instance data dictionary.

        If the field has a checksum field specified and this has not been initialised, then this one as well get
        assigned the same value.

        :param instance: a Model instance.
        :param value: the value to be assigned to the Field.
        """
        # do whatever is needed by the normal FieldAccessor
        super().__set__(instance, value)
        # if there is a linked field, continue here; otherwise, you have already finished.
        if self.field.checksum_field is not None:
            # check if the model has an attribute to store the initialisation of the linked field,
            # if not, create one and set it to False
            if not hasattr(instance, 'init_' + self.field.checksum_field):
                setattr(instance, 'init_' + self.field.checksum_field, False)

            # if the linked field has not been initialised, then set its value to the same value
            # of this field.
            if not getattr(instance, 'init_' + self.field.checksum_field):
                instance.__data__[self.field.checksum_field] = value
                instance._dirty.add(self.field.checksum_field)


# noinspection PyProtectedMember
class FileChecksumFieldAccessor(FieldAccessor):
    """
    Field accessor specialized for file checksum fields.

    When the field is directly set, then an initialization flag in the model instance is turned to True to avoid
    that the linked primary field will overrule this value again.

    For each checksum field named my_checksum, the model instance will get an attribute: init_my_checksum to be
    used as an initialization flag.

    Once the field is manually set, to re-establish the automatic mechanism, the user has to manually toggle the
    initialization flag.
    """

    def __set__(self, instance: Any, value: Any) -> None:
        """
        Sets the value of the field in the instance data dictionary.

        When the field is directly set, then the initialisation flag in the instance is also turned to True to avoid
        that the primary field will overrule this value again.

        For each checksum field named my_checksum, the model instance will get an attribute: init_my_checksum to be
        used as an initialisation flag.

        Once the field is manually set, to re-establish the automatic mechanism, the user has to manually toggle the
        initialisation flag.

        :param instance: a Model instance.
        :param value: the value to be assigned to the Field.
        """
        # do whatever is needed.
        super().__set__(instance, value)
        # check if the model has an attribute to store the initialisation of this field.
        # if not, create one and set it to False
        if not hasattr(instance, 'init_' + self.name):
            setattr(instance, 'init_' + self.name, False)
        # the field has been initialised, so set it to True, to avoid the primary field to overrule it.
        setattr(instance, 'init_' + self.name, True)


class FileNameField(TextField):
    """
    Field to be used for filenames.

    It is just an overload of TextField, that allows to apply filters and python functions specific to filenames.

    If the user specifies the name of a file checksum field, then when this field is updated, the checksum one will
    also be automatically updated.

    .. seealso::

        * :class:`~mafw.db.fields.FileNameListField` for a field able to store a list of filenames.

        * :func:`~mafw.tools.file_tools.remove_widow_db_rows` for a function removing entries from a database table
          where the corresponding files on disk are missing.

        * :func:`~mafw.tools.file_tools.verify_checksum` for a function comparing the actual checksum with the
          stored one and in case removing outdated entries from the DB.

    """

    accessor_class = FileNameFieldAccessor
    """The specific accessor class"""

    def __init__(self, checksum_field: str | None = None, *args: Any, **kwargs: Any) -> None:
        """
        Constructor parameter:

        :param checksum_field: The name of the checksum field linked to this filename. Defaults to None.
        :type checksum_field: str, Optional
        """
        super().__init__(*args, **kwargs)
        self.checksum_field = checksum_field

    def db_value(self, value: str | Path) -> str:
        """Converts the input python value into a string for the DB."""
        return str(value)

    def python_value(self, value: str) -> Path | None:
        """Converts the db value from str to Path

        The return value might also be None, if the user set the field value to null.

        :param value: The value of the field as stored in the database.
        :type value: str
        :return: The converted value as a path. It can be None, if value was stored as null.
        :rtype: Path | None
        """
        return Path(value) if value is not None else None


class FileNameListField(FileNameField):
    """
    A field for a list of file names.

    The evolution of the :class:`~mafw.db.fields.FileNameField`, this field is able to store a list of filenames as a
    ';' separated string of full paths.

    It is meant to be used when a processor is saving a bunch of correlated files that are to be used all together.

    In a similar way as its parent class, it can be link to a checksum field, in this case, the checksum of the whole
    file list will be calculated.
    """

    def db_value(self, value: list[str | Path] | str | Path) -> str:
        """Converts the list of paths in a ';' separated string"""
        if isinstance(value, (str, Path)):
            value = [value]
        return ';'.join([str(v) for v in value])

    def python_value(self, value: str) -> list[Path]:  # type: ignore[override]
        """Converts the ';' separated string in a list of paths"""
        if value is None:  # this might be the case, when the database field is actually set to null.
            return []
        return [Path(p) for p in value.split(';')]


class FileChecksumField(TextField):
    """
    A field to be used for file checksum.

    It is the evolution of the TextField for storing file checksum hexadecimal digest.

    If linked to a :class:`FileNameField` or :class:`FileNameListField`, then it will be automatically filled when the
    primary file name (or list of file names) field is set.

    If the user decides to set its value manually, then he can provide either the string with the hexadecimal
    characters as calculated by :func:`~mafw.tools.file_tools.file_checksum`, or simply the filename (or filename
    list) and the field will perform the calculation automatically.

    """

    accessor_class = FileChecksumFieldAccessor
    """The specific field accessor class"""

    def db_value(self, value: str | Path | list[str | Path]) -> str:
        """
        Converts the python assigned value to the DB type.

        The checksum will be stored in the DB as a string containing only hexadecimal characters
        (see `hexdigest <https://docs.python.org/3/library/hashlib.html#hashlib.hash.hexdigest>`_).

        The user can provide the checksum directly or the path to the file or a list of path to files. If a Path,
        or list of Path,  is provided, then the checksum will be calculated, if a str (a path converted into a
        string) is provided, the function will try to see if a file with that path exists. If so, the checksum will
        be calculated, if not, the original string is assumed to be the checksum.

        :param value: The checksum or path to the file, or list of path to files for which the checksum has to be
            stored.
        :type value: str | Path | list[str | Path]
        :return: The checksum string for the database storage.
        :rtype: str
        """

        if isinstance(value, Path):
            # we got the filename, not the digest. we need to calculate it
            value = mafw.tools.file_tools.file_checksum(value)
        elif isinstance(value, list):
            # we got a list of path, not the digest. we need to calculate the digest
            # of the whole list
            value = mafw.tools.file_tools.file_checksum(value)
        else:
            test_value = Path(value)
            if test_value.exists():
                # with a very high probability, the user passed a str to a path.
                value = mafw.tools.file_tools.file_checksum(test_value)
        return value

    def python_value(self, value: str) -> str:
        return value



#  Copyright 2025 European Union
#  Author: Bulgheroni Antonio (antonio.bulgheroni@ec.europa.eu)
#  SPDX-License-Identifier: EUPL-1.2
"""
Module implements some basic functions involving regular expressions.
"""

import re


def extract_protocol(url: str) -> str | None:
    """
    Extract the protocol portion from a database connection URL.

    The extract_protocol function takes a database connection URL string as input and extracts the protocol portion
    (the part before "://"). This function is useful for identifying the database type from connection strings.

    :param url: The url from which the protocol will be extracted.
    :type url: str
    :return: The protocol or None, if the extraction failed
    :rtype: str | None
    """
    pattern = r'^([a-z0-9_\-+.]+)://'
    match = re.match(pattern, url)
    if match:
        return match.group(1)
    return None


def normalize_sql_spaces(sql_string: str) -> str:
    """
    Normalize multiple consecutive spaces in SQL string to single spaces.
    Only handles spaces, preserves other whitespace characters.

    :param sql_string: The SQL string for space normalization.
    :type sql_string: str
    :return: The normalized SQL command.
    :rtype: str
    """
    return re.sub(r' +', ' ', sql_string.strip())



#  Copyright 2025 European Union
#  Author: Bulgheroni Antonio (antonio.bulgheroni@ec.europa.eu)
#  SPDX-License-Identifier: EUPL-1.2
from peewee import SQL, AutoField, FloatField, ForeignKeyField, TextField

from mafw.db.db_model import MAFwBaseModel
from mafw.db.fields import FileChecksumField, FileNameField
from mafw.db.std_tables import StandardTable
from mafw.db.trigger import Trigger, TriggerAction, TriggerWhen, or_


class Detector(StandardTable):
    detector_id = AutoField(primary_key=True, help_text='Primary key for the detector table')
    name = TextField(help_text='The name of the detector')
    description = TextField(help_text='A longer description for the detector')

    @classmethod
    def init(cls) -> None:
        data = [
            dict(detector_id=1, name='Normal', description='Standard detector'),
            dict(detector_id=2, name='HighGain', description='High gain detector'),
            dict(detector_id=3, name='NoDark', description='Low dark current detector'),
        ]

        cls.insert_many(data).on_conflict(
            conflict_target=[cls.detector_id],
            update={'name': SQL('EXCLUDED.name'), 'description': SQL('EXCLUDED.description')},
        ).execute()


class InputFile(MAFwBaseModel):
    @classmethod
    def triggers(cls) -> list[Trigger]:
        update_file_trigger = Trigger(
            trigger_name='input_file_after_update',
            trigger_type=(TriggerWhen.After, TriggerAction.Update),
            source_table=cls,
            safe=True,
            for_each_row=True,
        )
        update_file_trigger.add_when(or_('NEW.exposure != OLD.exposure', 'NEW.checksum != OLD.checksum'))
        update_file_trigger.add_sql('DELETE FROM data WHERE file_pk = OLD.file_pk;')

        return [update_file_trigger]

    file_pk = AutoField(primary_key=True, help_text='Primary key for the input file table')
    filename = FileNameField(unique=True, checksum_field='checksum', help_text='The filename of the element')
    checksum = FileChecksumField(help_text='The checksum of the element file')
    exposure = FloatField(help_text='The duration of the exposure in h')
    detector = ForeignKeyField(
        Detector, Detector.detector_id, on_delete='CASCADE', backref='detector', column_name='detector_id'
    )
    # finish input file


class Data(MAFwBaseModel):
    @classmethod
    def triggers(cls) -> list[Trigger]:
        delete_plotter_sql = 'DELETE FROM plotter_output WHERE plotter_name = "PlugPlotter";'

        insert_data_trigger = Trigger(
            trigger_name='data_after_insert',
            trigger_type=(TriggerWhen.After, TriggerAction.Insert),
            source_table=cls,
            safe=True,
            for_each_row=False,
        )
        insert_data_trigger.add_sql(delete_plotter_sql)

        update_data_trigger = Trigger(
            trigger_name='data_after_update',
            trigger_type=(TriggerWhen.After, TriggerAction.Update),
            source_table=cls,
            safe=True,
            for_each_row=False,
        )
        update_data_trigger.add_when('NEW.value != OLD.value')
        update_data_trigger.add_sql(delete_plotter_sql)

        delete_data_trigger = Trigger(
            trigger_name='data_after_delete',
            trigger_type=(TriggerWhen.After, TriggerAction.Delete),
            source_table=cls,
            safe=True,
            for_each_row=False,
        )
        delete_data_trigger.add_sql(delete_plotter_sql)

        return [insert_data_trigger, delete_data_trigger, update_data_trigger]

    file_pk = ForeignKeyField(InputFile, on_delete='cascade', backref='file', primary_key=True, column_name='file_id')
    value = FloatField(help_text='The result of the measurement')



#  Copyright 2025 European Union
#  Author: Bulgheroni Antonio (antonio.bulgheroni@ec.europa.eu)
#  SPDX-License-Identifier: EUPL-1.2
"""
The module provides some general decorator utilities that are used in several parts of the code, and that can be
reused by the user community.
"""

import functools
import typing
import warnings
from importlib.util import find_spec
from typing import Any, Callable, Type

from mafw.enumerators import LoopType
from mafw.mafw_errors import MissingDatabase, MissingOptionalDependency
from mafw.processor import Processor

F = typing.TypeVar('F', bound=typing.Callable[..., object])
"""TypeVar for generic function."""

# Define a TypeVar to capture Processor subclasses
P = typing.TypeVar('P', bound=Processor)
"""TypeVar for generic processor."""


def suppress_warnings(func: F) -> F:
    """
    Decorator to suppress warnings during the execution of a test function.

    This decorator uses the `warnings.catch_warnings()` context manager to
    temporarily change the warning filter to ignore all warnings. It is useful
    when you want to run a test without having warnings clutter the output.

    Usage::

        @suppress_warnings
        def test_function():
            # Your test code that might emit warnings


    :param func: The test function to be decorated.
    :type func: Callable
    :return: The wrapped function with suppressed warnings.
    :rtype: Callable
    """

    @functools.wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> F:
        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            return func(*args, **kwargs)  # type: ignore[return-value]

    return wrapper  # type: ignore[return-value]


@typing.no_type_check  # no idea how to fix it
def singleton(cls):
    """Make a class a Singleton class (only one instance)"""

    @functools.wraps(cls)
    def wrapper_singleton(*args: Any, **kwargs: Any):
        if wrapper_singleton.instance is None:
            wrapper_singleton.instance = cls(*args, **kwargs)
        return wrapper_singleton.instance

    wrapper_singleton.instance = None
    return wrapper_singleton


@typing.no_type_check
def database_required(cls):
    """Modify the processor start method to check if a database object exists.

    This decorator must be applied to processors requiring a database connection.

    :param cls: A Processor class.
    """
    orig_start = cls.start

    @functools.wraps(cls.start)
    def _start(self) -> None:
        if self._database is None:
            raise MissingDatabase(f'{self.name} requires an active database.')
        orig_start(self)

    cls.start = _start

    return cls


@typing.no_type_check
def orphan_protector(cls):
    """
    A class decorator to modify the init method of a Processor so that the remove_orphan_files is set to False and
    no orphan files will be removed.
    """
    old_init = cls.__init__

    @functools.wraps(cls.__init__)
    def new_init(self, *args, **kwargs):
        old_init(self, *args, remove_orphan_files=False, **kwargs)

    cls.__init__ = new_init
    return cls


@typing.no_type_check
def execution_workflow(loop_type: LoopType | str = LoopType.ForLoop):
    """
    A decorator factory for the definition of the looping strategy.

    This decorator factory must be applied to Processor subclasses to modify their value of loop_type in order to
    change the execution workflow.

    See :func:`single_loop`, :func:`for_loop` and :func:`while_loop` decorator shortcuts.

    :param loop_type: The type of execution workflow requested for the decorated class. Defaults to LoopType.ForLoop.
    :type loop_type: LoopType | str, Optional
    """

    def dec(cls):
        """The class decorator."""
        old_init = cls.__init__

        @functools.wraps(cls.__init__)
        def new_init(self, *args, **kwargs):
            """The modified Processor init"""
            old_init(self, *args, looper=loop_type, **kwargs)

        cls.__init__ = new_init

        return cls

    return dec


single_loop = execution_workflow(LoopType.SingleLoop)
"""A decorator shortcut to define a single execution processor."""

for_loop = execution_workflow(LoopType.ForLoop)
"""A decorator shortcut to define a for loop execution processor."""

while_loop = execution_workflow(LoopType.WhileLoop)
"""A decorator shortcut to define a while loop execution processor."""


def depends_on_optional(
    module_name: str, raise_ex: bool = False, warn: bool = True
) -> Callable[[F], Callable[..., Any]]:
    """
    Function decorator to check if module_name is available.

    If module_name is found, then returns the wrapped function. If not, its behavior depends on the raise_ex and
    warn_only values. If raise_ex is True, then an ImportError exception is raised. If it is False and warn is
    True, then a warning message is displayed but no exception is raised. If they are both False, then function is
    silently skipped.

    If raise_ex is True, the value of `warn` is not taken into account.

    **Typical usage**

    The user should decorate functions or class methods when they cannot be executed without the optional library.
    In the specific case of Processor subclass, where the class itself can be created also without the missing
    library, but it is required somewhere in the processor execution, then the user is suggested to decorate the
    execute method with this decorator.

    :param module_name: The optional module(s) from which the function depends on. A ";" separated list of modules can
        also be provided.
    :type module_name: str
    :param raise_ex: Flag to raise an exception if module_name is not found, defaults to False.
    :type raise_ex: bool, Optional
    :param warn: Flag to display a warning message if module_name is not found, default to True.
    :type warn: bool, Optional
    :return: The wrapped function
    :rtype: Callable
    :raise ImportError: if module_name is not found and raise_ex is True.
    """

    def decorator(func: F) -> Callable[..., Any]:
        @functools.wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            all_mods_found = all(find_spec(mod.strip()) is not None for mod in module_name.split(';'))
            if not all_mods_found:
                msg = f'Optional dependency {module_name} not found ({func.__qualname__})'
                if raise_ex:
                    raise ImportError(msg)
                else:
                    if warn:
                        warnings.warn(MissingOptionalDependency(msg), stacklevel=2)
                    return None  # Explicitly return None when skipping the function
            else:
                return func(*args, **kwargs)

        return wrapper

    return decorator


def processor_depends_on_optional(
    module_name: str, raise_ex: bool = False, warn: bool = True
) -> Callable[[Type[P]], Type[Processor]]:
    """
    Class decorator factory to check if module module_name is available.

    It checks if all the optional modules listed in `module_name` separated by a ';' can be found.

    If all modules are found, then the class is returned as it is.

    If at least one module is not found:
     - and raise_ex is True, an ImportError exception is raised and the user is responsible to deal with it.
     - if raise_ex is False, instead of returning the class, the :class:`~.Processor` is returned.
     - depending on the value of warn, the user will be informed with a warning message or not.

    **Typical usage**

    The user should decorate Processor subclasses everytime the optional module is required in their __init__ method.
    Should the check on the optional module have a positive outcome, then the Processor subclass is returned.
    Otherwise, if raise_ex is False, an instance of the base :py:class:`~.Processor` is returned. In
    this way, the returned class can still be executed without breaking the execution scheme but of course, without
    producing any output.

    Should be possible to run the __init__ method of the class without the missing library, then the user can also
    follow the approach described in this other :func:`example <depends_on_optional>`.

    :param module_name: The optional module(s) from which the class depends on. A ";" separated list of modules can
        also be provided.
    :type module_name: str
    :param raise_ex: Flag to raise an exception if module_name not found, defaults to False.
    :type raise_ex: bool, Optional
    :param warn: Flag to display a warning message if module_name is not found, defaults to True.
    :type warn: bool, Optional
    :return: The wrapped processor.
    :rtype: type(Processor)
    :raise ImportError: if module_name is not found and raise_ex is True.
    """

    def decorator(cls: Type[P]) -> Type[Processor]:
        """
        The class decorator.

        It checks if all the modules provided by the decorator factory are available on the systems.
        If yes, then it simply returns `cls`. If no, it returns a subclass of the :class:`~.Processor`
        after all the introspection properties have been taken from `cls`.

        :param cls: The class being decorated.
        :type cls: type(Processor)
        :return: The decorated class, either cls or a subclass of  :class:`~autorad.processor.Processor`.
        :rtype: type(Processor)
        """

        def class_wrapper(klass: Type[Processor]) -> Type[Processor]:
            """
            Copy introspection properties from cls to klass.

            :param klass: The class to be modified.
            :type klass: class.
            :return: The modified class.
            :rtype: class.
            """
            klass.__module__ = cls.__module__
            klass.__name__ = f'{cls.__name__} (Missing {module_name})'
            klass.__qualname__ = cls.__qualname__
            klass.__annotations__ = cls.__annotations__
            klass.__doc__ = cls.__doc__
            return klass

        all_mods_found = all([find_spec(mod.strip()) is not None for mod in module_name.split(';')])
        if not all_mods_found:
            msg = f'Optional dependency {module_name} not found ({cls.__qualname__})'
            if raise_ex:
                raise ImportError(msg)
            else:
                if warn:
                    warnings.warn(MissingOptionalDependency(msg), stacklevel=2)

                # We subclass the basic processor.
                class NewClass(Processor):
                    pass

                # The class wrapper is copying introspection properties from the cls to the NewClass
                new_class = class_wrapper(NewClass)

        else:
            new_class = cls
        return new_class

    return decorator


def class_depends_on_optional(
    module_name: str, raise_ex: bool = False, warn: bool = True
) -> Callable[[Type[Any]], Type[Any]]:
    """
    Class decorator factory to check if module module_name is available.

    It checks if all the optional modules listed in `module_name` separated by a ';' can be found.

    If all modules are found, then the class is returned as it is.

    If at least one module is not found:
     - and raise_ex is True, an ImportError exception is raised and the user is responsible to deal with it.
     - if raise_ex is False, instead of returning the class, a new empty class is returned.
     - depending on the value of warn, the user will be informed with a warning message or not.

    :param module_name: The optional module(s) from which the class depends on. A ";" separated list of modules can
        also be provided.
    :type module_name: str
    :param raise_ex: Flag to raise an exception if module_name not found, defaults to False.
    :type raise_ex: bool, Optional
    :param warn: Flag to display a warning message if module_name is not found, defaults to True.
    :type warn: bool, Optional
    :return: The wrapped class.
    :rtype: type(object)
    :raise ImportError: if module_name is not found and raise_ex is True.
    """

    def decorator(cls: Type[Any]) -> Type[Any]:
        """
        The class decorator.

        It checks if all the modules provided by the decorator factory are available on the systems.
        If yes, then it simply returns `cls`. If no, it returns a subclass of the cls bases.
        after all the introspection properties have been taken from `cls`.

        :param cls: The class being decorated.
        :type cls: type(cls)
        :return: The decorated class, either cls or a subclass of cls.
        :rtype: type(cls)
        """

        def class_wrapper(klass: Type[Any]) -> Type[Any]:
            """
            Copy introspection properties from cls to klass.

            :param klass: The class to be modified.
            :type klass: class.
            :return: The modified class.
            :rtype: class.
            """
            klass.__module__ = cls.__module__
            klass.__name__ = f'{cls.__name__} (Missing {module_name})'
            klass.__qualname__ = cls.__qualname__
            klass.__annotations__ = cls.__annotations__
            klass.__doc__ = cls.__doc__
            return klass

        all_mods_found = all([find_spec(mod.strip()) is not None for mod in module_name.split(';')])
        if not all_mods_found:
            msg = f'Optional dependency {module_name} not found ({cls.__qualname__})'
            if raise_ex:
                raise ImportError(msg)
            else:
                if warn:
                    warnings.warn(MissingOptionalDependency(msg), stacklevel=2)

                # we subclass the original class.
                class NewClass(*cls.__bases__):  # type: ignore
                    pass

                # the class wrapper is copying introspection properties from the cls to the NewClass
                new_class = class_wrapper(NewClass)

        else:
            new_class = cls
        return new_class

    return decorator



#  Copyright 2025 European Union
#  Author: Bulgheroni Antonio (antonio.bulgheroni@ec.europa.eu)
#  SPDX-License-Identifier: EUPL-1.2
"""
Unit tests for the console_user_interface module.

This module contains comprehensive tests for the ConsoleInterface class,
focusing on its logging-based functionality and inherited behavior.
"""

import logging
from unittest.mock import Mock, patch

import pytest

from mafw.enumerators import ProcessorStatus
from mafw.ui.abstract_user_interface import UserInterfaceBase
from mafw.ui.console_user_interface import ConsoleInterface


class TestConsoleInterface:
    """Test class for ConsoleInterface functionality."""

    def test_inheritance(self):
        """Test that ConsoleInterface properly inherits from UserInterfaceBase."""
        assert issubclass(ConsoleInterface, UserInterfaceBase)

        interface = ConsoleInterface()
        assert isinstance(interface, UserInterfaceBase)
        assert isinstance(interface, ConsoleInterface)

    def test_class_attributes(self):
        """Test that ConsoleInterface has the correct class attributes."""
        assert hasattr(ConsoleInterface, 'name')
        assert ConsoleInterface.name == 'console'

        # Should inherit from base class
        assert hasattr(ConsoleInterface, 'always_display_progress_message')
        assert ConsoleInterface.always_display_progress_message == 10

    def test_instantiation(self):
        """Test that ConsoleInterface can be instantiated."""
        interface = ConsoleInterface()
        assert interface.name == 'console'
        assert interface.always_display_progress_message == 10

    def test_create_task_method_signature(self):
        """Test create_task method accepts all expected parameters."""
        interface = ConsoleInterface()

        # Should not raise any exceptions (method just passes)
        interface.create_task('test_task')
        interface.create_task('test_task', 'description')
        interface.create_task('test_task', 'description', completed=50)
        interface.create_task('test_task', 'description', completed=50, increment=10)
        interface.create_task('test_task', 'description', completed=50, increment=10, total=100)
        interface.create_task('test_task', extra_kwarg='value')

    def test_update_task_method_signature(self):
        """Test update_task method accepts all expected parameters."""
        interface = ConsoleInterface()

        # Should not raise any exceptions (method just passes)
        interface.update_task('test_task')
        interface.update_task('test_task', completed=50)
        interface.update_task('test_task', completed=50, increment=10)
        interface.update_task('test_task', completed=50, increment=10, total=100)
        interface.update_task('test_task', extra_kwarg='value')

    def test_context_manager_enter(self):
        """Test __enter__ method returns self."""
        interface = ConsoleInterface()
        result = interface.__enter__()
        assert result is interface

    @pytest.mark.parametrize(
        'exception_type,exception_value,traceback_value',
        [
            (None, None, None),
            (ValueError, ValueError('test error'), None),
            (RuntimeError, RuntimeError('runtime error'), Mock()),
        ],
    )
    def test_context_manager_exit(self, exception_type, exception_value, traceback_value):
        """Test __exit__ method handles various scenarios."""
        interface = ConsoleInterface()

        # Should not raise any exceptions and return None
        result = interface.__exit__(exception_type, exception_value, traceback_value)
        assert result is None

    def test_context_manager_usage(self):
        """Test full context manager usage."""
        interface = ConsoleInterface()

        with interface as ctx:
            assert ctx is interface
            # Should be able to call methods within context
            ctx.create_task('test')
            ctx.update_task('test')


class TestConsoleInterfaceDisplayProgressMessage:
    """Test class for display_progress_message functionality."""

    @pytest.fixture
    def interface(self):
        """Fixture providing ConsoleInterface instance."""
        return ConsoleInterface()

    @pytest.fixture
    def mock_logger(self):
        """Fixture providing mocked logger."""
        with patch('mafw.ui.console_user_interface.log') as mock_log:
            yield mock_log

    def test_display_progress_message_time_check_delegation(self, interface, mock_logger):
        """Test that display_progress_message delegates time check to parent method."""
        with patch.object(interface, '_is_time_to_display_lopping_message') as mock_time_check:
            mock_time_check.return_value = False

            interface.display_progress_message('Test message', 5, 100, 0.1)

            mock_time_check.assert_called_once_with(5, 100, 0.1)
            mock_logger.info.assert_not_called()

    @pytest.mark.parametrize(
        'i_item,n_item,frequency,expected_calls',
        [
            (0, 100, 0.1, 1),  # First item should display
            (10, 100, 0.1, 1),  # At frequency interval
            (99, 100, 0.1, 1),  # Last item should display
            (5, 100, 0.1, 0),  # Not at frequency interval
            (0, 5, 0.1, 1),  # Small n_item, always display
            (2, 5, 0.1, 1),  # Small n_item, always display
        ],
    )
    def test_display_progress_message_logging_frequency(
        self, interface, mock_logger, i_item, n_item, frequency, expected_calls
    ):
        """Test logging frequency based on inherited logic."""
        interface.display_progress_message('Test message', i_item, n_item, frequency)

        assert mock_logger.info.call_count == expected_calls

    def test_display_progress_message_with_none_n_item(self, interface, mock_logger):
        """Test display_progress_message behavior when n_item is None."""
        # Should use max(1000, i_item) as n_item for display purposes
        interface.display_progress_message('Processing item', 500, None, 0.1)

        # Should log since 500 is multiple of 10 (inherited logic for None n_item)
        mock_logger.info.assert_called()  # 500 % 10 != 0

        # Test with item that should display
        interface.display_progress_message('Processing item', 10, None, 0.1)
        assert len(mock_logger.info.call_args_list) == 2

    def test_display_progress_message_formatting_with_known_total(self, interface, mock_logger):
        """Test message formatting when total is known."""
        interface.display_progress_message('Processing files', 0, 100, 0.1)  # First item always displays

        mock_logger.info.assert_called_once()
        call_args = mock_logger.info.call_args[0][0]

        # Should contain counter format and message
        assert '[  1/100]' in call_args
        assert 'Processing files' in call_args

    def test_display_progress_message_formatting_with_none_total(self, interface, mock_logger):
        """Test message formatting when total is None."""
        interface.display_progress_message('Processing items', 0, None, 0.1)

        mock_logger.info.assert_called_once()
        call_args = mock_logger.info.call_args[0][0]

        # When n_item is None, it should use max(1000, i_item) = max(1000, 0) = 1000
        assert '[   1/1000]' in call_args
        assert 'Processing items' in call_args

    def test_display_progress_message_formatting_with_large_i_item_and_none_total(self, interface, mock_logger):
        """Test message formatting when i_item > 1000 and total is None."""
        interface.display_progress_message('Processing items', 2000, None, 0.1)

        # 2000 % 10 == 0, so should display
        mock_logger.info.assert_called_once()
        call_args = mock_logger.info.call_args[0][0]

        # When n_item is None and i_item > 1000, should use i_item as total
        assert '[2001/2000]' in call_args
        assert 'Processing items' in call_args

    @pytest.mark.parametrize(
        'i_item,n_item,expected_width,expected_counter',
        [
            (0, 10, 2, '[ 1/10]'),
            (99, 100, 3, '[100/100]'),
            (999, 1000, 4, '[1000/1000]'),
            (0, 1, 1, '[1/1]'),
        ],
    )
    def test_display_progress_message_counter_formatting(
        self, interface, mock_logger, i_item, n_item, expected_width, expected_counter
    ):
        """Test counter formatting with different widths."""
        interface.display_progress_message('Test', i_item, n_item, 1.0)  # High frequency to ensure display

        if mock_logger.info.called:
            call_args = mock_logger.info.call_args[0][0]
            assert expected_counter in call_args

    def test_display_progress_message_counter_width_calculation(self, interface, mock_logger):
        """Test that counter width is calculated correctly."""
        # Test with 4-digit number to verify width calculation
        interface.display_progress_message('Test', 0, 1000, 1.0)

        mock_logger.info.assert_called_once()
        call_args = mock_logger.info.call_args[0][0]

        # Should be right-aligned to width of str(1000) = 4
        assert '[   1/1000]' in call_args

    def test_display_progress_message_complete_format(self, interface, mock_logger):
        """Test complete message format including counter and message."""
        message = 'Processing data files'
        interface.display_progress_message(message, 0, 50, 1.0)

        mock_logger.info.assert_called_once()
        call_args = mock_logger.info.call_args[0][0]

        expected = f'[ 1/50] {message}'
        assert call_args == expected


class TestConsoleInterfaceChangeOfProcessorStatus:
    """Test class for change_of_processor_status functionality."""

    @pytest.fixture
    def interface(self):
        """Fixture providing ConsoleInterface instance."""
        return ConsoleInterface()

    @pytest.fixture
    def mock_logger(self):
        """Fixture providing mocked logger."""
        with patch('mafw.ui.console_user_interface.log') as mock_log:
            yield mock_log

    def test_change_of_processor_status_logs_debug_message(self, interface, mock_logger):
        """Test that processor status change logs debug message."""
        processor_name = 'test_processor'
        old_status = ProcessorStatus.Init
        new_status = ProcessorStatus.Start

        interface.change_of_processor_status(processor_name, old_status, new_status)

        mock_logger.debug.assert_called_once()
        call_args = mock_logger.debug.call_args[0][0]

        expected_message = f'{processor_name} is {new_status}'
        assert call_args == expected_message

    @pytest.mark.parametrize(
        'processor_name,old_status,new_status',
        [
            ('processor_1', ProcessorStatus.Init, ProcessorStatus.Start),
            ('data_loader', ProcessorStatus.Start, ProcessorStatus.Run),
            ('file_processor', ProcessorStatus.Run, ProcessorStatus.Finish),
            ('validator', ProcessorStatus.Finish, ProcessorStatus.Init),
        ],
    )
    def test_change_of_processor_status_various_statuses(
        self, interface, mock_logger, processor_name, old_status, new_status
    ):
        """Test processor status change with various status combinations."""
        interface.change_of_processor_status(processor_name, old_status, new_status)

        mock_logger.debug.assert_called_once()
        call_args = mock_logger.debug.call_args[0][0]

        expected_message = f'{processor_name} is {new_status}'
        assert call_args == expected_message

    def test_change_of_processor_status_message_format(self, interface, mock_logger):
        """Test the exact format of the status change message."""
        processor_name = 'my_custom_processor'
        old_status = ProcessorStatus.Init
        new_status = ProcessorStatus.Run

        interface.change_of_processor_status(processor_name, old_status, new_status)

        # Verify the exact message format
        expected_message = f'{processor_name} is {new_status}'
        mock_logger.debug.assert_called_once_with(expected_message)

    def test_change_of_processor_status_ignores_old_status(self, interface, mock_logger):
        """Test that old_status is ignored in the log message."""
        processor_name = 'test_processor'
        old_status = ProcessorStatus.Finish
        new_status = ProcessorStatus.Init

        interface.change_of_processor_status(processor_name, old_status, new_status)

        call_args = mock_logger.debug.call_args[0][0]

        # Message should only contain processor name and new status
        assert processor_name in call_args
        assert str(new_status) in call_args
        assert str(old_status) not in call_args


@pytest.mark.integration_test
class TestConsoleInterfaceIntegration:
    """Integration tests for ConsoleInterface."""

    @pytest.fixture
    def interface(self):
        """Fixture providing ConsoleInterface instance."""
        return ConsoleInterface()

    def test_logger_module_name(self):
        """Test that the logger has the correct module name."""
        from mafw.ui.console_user_interface import log

        # Verify logger has correct name (it's created at import time)
        assert log.name == 'mafw.ui.console_user_interface'

    def test_full_workflow_with_logging(self, interface):
        """Test complete workflow with all logging methods."""
        with patch('mafw.ui.console_user_interface.log') as mock_logger:
            # Use as context manager
            with interface as ctx:
                # Create and update tasks (no logging expected)
                ctx.create_task('test_task', 'Test description', total=100)
                ctx.update_task('test_task', completed=50)

                # Display progress messages
                ctx.display_progress_message('Processing', 0, 10, 0.1)  # Should log (first item)
                ctx.display_progress_message('Processing', 5, 10, 0.1)  # Should log (small total)

                # Change processor status
                ctx.change_of_processor_status('processor', ProcessorStatus.Init, ProcessorStatus.Run)

            # Verify logging calls
            assert mock_logger.info.call_count >= 1  # At least one progress message
            mock_logger.debug.assert_called_once()  # Status change

    def test_interface_satisfies_metaclass_requirements(self, interface):
        """Test that ConsoleInterface satisfies all metaclass requirements."""
        # This should pass since ConsoleInterface inherits from UserInterfaceBase
        assert isinstance(interface, UserInterfaceBase)

        # Verify all required methods are callable
        required_methods = ['create_task', 'update_task', 'display_progress_message', 'change_of_processor_status']
        for method in required_methods:
            assert hasattr(interface, method)
            assert callable(getattr(interface, method))

    def test_console_interface_docstring_and_class_attributes(self):
        """Test docstring and class-level attributes."""
        assert ConsoleInterface.__doc__ is not None
        assert 'console user interface' in ConsoleInterface.__doc__.lower()
        assert 'headless' in ConsoleInterface.__doc__.lower()

        # Test class attribute
        assert ConsoleInterface.name == 'console'

    def test_logging_integration_with_real_logger(self, interface):
        """Test integration with actual logging system."""
        # Create a test logger to capture output
        test_logger = logging.getLogger('test_console_interface')
        test_logger.setLevel(logging.DEBUG)

        # Create a list handler to capture log records
        log_records = []

        class ListHandler(logging.Handler):
            def emit(self, record):
                log_records.append(record)

        handler = ListHandler()
        test_logger.addHandler(handler)

        # Patch the module logger to use our test logger
        with patch('mafw.ui.console_user_interface.log', test_logger):
            interface.display_progress_message('Test message', 0, 100, 0.1)
            interface.change_of_processor_status('test_proc', ProcessorStatus.Init, ProcessorStatus.Run)

        # Verify log records
        assert len(log_records) == 2

        # Check info log (progress message)
        info_record = next(r for r in log_records if r.levelno == logging.INFO)
        assert 'Test message' in info_record.getMessage()
        assert '[  1/100]' in info_record.getMessage()

        # Check debug log (status change)
        debug_record = next(r for r in log_records if r.levelno == logging.DEBUG)
        assert f'test_proc is {ProcessorStatus.Run}' in debug_record.getMessage()

    @pytest.mark.parametrize(
        'method_name,args,kwargs',
        [
            ('create_task', ('task1',), {'task_description': 'desc', 'total': 100}),
            ('update_task', ('task1',), {'completed': 50}),
            ('display_progress_message', ('msg', 1, 10, 0.1), {}),
            ('change_of_processor_status', ('proc', ProcessorStatus.Init, ProcessorStatus.Run), {}),
        ],
    )
    def test_method_calls_dont_raise_exceptions(self, interface, method_name, args, kwargs):
        """Test that all methods can be called without raising exceptions."""
        method = getattr(interface, method_name)

        # Should not raise any exceptions
        try:
            method(*args, **kwargs)
        except Exception as e:
            pytest.fail(f'{method_name} raised an unexpected exception: {e}')



#  Copyright 2025 European Union
#  Author: Bulgheroni Antonio (antonio.bulgheroni@ec.europa.eu)
#  SPDX-License-Identifier: EUPL-1.2
import pytest

from mafw.active import Active


class TestActive:
    def setup_method(self):
        class TestClass:
            variable = Active(default=0)

            def on_variable_change(self, old_value, new_value):
                self.change_called = True
                self.old_value = old_value
                self.new_value = new_value

            def on_variable_set(self, value):
                self.set_called = True
                self.set_value = value

            def on_variable_get(self, value):
                self.get_called = True
                self.get_value = value

        self.test_obj = TestClass()

    @pytest.mark.parametrize(
        'initial, new, expected_change_call',
        [
            (0, 1, True),
            (1, 1, False),
            (5, 10, True),
        ],
    )
    def test_change_callback(self, initial, new, expected_change_call):
        self.test_obj.variable = initial
        self.test_obj.change_called = False
        self.test_obj.variable = new
        assert self.test_obj.change_called == expected_change_call

    @pytest.mark.parametrize(
        'initial, new, expected_set_call',
        [
            (0, 1, False),
            (1, 1, True),
            (5, 5, True),
        ],
    )
    def test_set_callback(self, initial, new, expected_set_call):
        self.test_obj.variable = initial
        self.test_obj.set_called = False
        self.test_obj.variable = new
        assert self.test_obj.set_called == expected_set_call

    @pytest.mark.parametrize('initial', [0, 1, 10])
    def test_get_callback(self, initial):
        self.test_obj.variable = initial
        self.test_obj.get_called = False
        _ = self.test_obj.variable
        assert self.test_obj.get_called

    def test_default_value(self):
        assert self.test_obj.variable == 0



#  Copyright 2025 European Union
#  Author: Bulgheroni Antonio (antonio.bulgheroni@ec.europa.eu)
#  SPDX-License-Identifier: EUPL-1.2
"""
The module provides the capability to filter model classes with information taken from the steering file.
"""

import operator
import warnings
from collections import UserDict
from copy import copy
from functools import reduce
from typing import Any, Self, cast

import peewee
from peewee import Model


class Filter:
    r"""
    Class to filter rows from a model.

    The filter object can be used to generate a where clause to be applied to Model.select().

    The construction of a Filter is normally done via a configuration file using the :meth:`from_conf` class method.
    The name of the filter is playing a key role in this. If it follows a dot structure like:

        *ProcessorName.Filter.ModelName*

    then the corresponding table from the TOML configuration object will be used.

    For each processor, there might be many Filters, up to one for each Model used to get the input list. If a
    processor is joining together three Models when performing the input select, there will be up to three Filters
    collaborating on making the selection.

    The filter configuration can contain the following key, value pair:

        - key / string pairs, where the key is the name of a field in the corresponding Model

        - key / numeric pairs

        - key / arrays

    All fields from the configuration file will be added to the instance namespace, thus accessible with the dot
    notation. Moreover, the field names and their filter value will be added to a private dictionary to simplify the
    generation of the filter SQL code.

    The user can use the filter object to store selection criteria. He can construct queries using the filter
    contents in the same way as he could use processor parameters.

    If he wants to automatically generate valid filtering expression, he can use the :meth:`filter` method. In order
    for this to work, the Filter object be :meth:`bound <bind>` to a Model. Without this binding the Filter will not
    be able to automatically generate expressions.

    For each field in the filter, one condition will be generated according to the following scheme:

    =================   =================   ==================
    Filter field type   Logical operation      Example
    =================   =================   ==================
    Numeric, boolean        ==               Field == 3.14
    String                 GLOB             Field GLOB '\*ree'
    List                   IN               Field IN [1, 2, 3]
    =================   =================   ==================

    All conditions will be joined with a AND logic.

    Consider the following example:

    .. code-block:: python
        :linenos:

        class MeasModel(MAFwBaseModel):
            meas_id = AutoField(primary_key=True)
            sample_name = TextField()
            successful = BooleanField()

        flt = Filter('MyProcessor.Filter.MyModel', sample_name='sample_00*',
                    meas_id=[1,2,3], successful=True)
        flt.bind(MeasModel)

        filter_select_manual = MeasModel.select().where((MeasModel.meas_id.in_(flt.meas_id) &
                                                        (MeasModel.sample_name % flt.sample_name) &
                                                        (MeasModel.successful == flt.successful))

        filter_select_auto = MeasModel.select().where(flt.filter())

    At line 6, we created a Filter and at 8 we bind it to our Model class. The two select at 10 and 14 are actually
    identical.

    """

    def __init__(self, name_: str, **kwargs: Any) -> None:
        """
        Constructor parameters:

        :param `name_`: The name of the filter. It should be in dotted format to facilitate the configuration via the
            steering file. The _ is used to allow the user to have a keyword argument named name.
        :type `name_`: str
        :param kwargs: Keyword parameters corresponding to fields and filter values.

        .. versionchanged:: 1.2.0
            The parameter *name* has been renamed as *name_*.

        """
        self.name = name_
        self.model_name = name_.split('.')[-1]
        self.model: type[Model] | None = None
        self._model_bound: bool = False
        self._fields = {}
        for k, v in kwargs.items():
            self._fields[k] = v
            setattr(self, k, v)

    def bind(self, model: type[Model] | None = None) -> None:
        """
        Connects a filter to a Model class.

        If no model is provided, the method will try to bind a class from with global dictionary with a name matching
        the model name used during the Filter configuration. It only works when the Model is defined as global.

        :param model: Model to be bound. Defaults to None
        :type model: Model, Optional
        """
        if model is None:
            if self.model_name in globals():
                self.model = globals()[self.model_name]
                self._model_bound = True
        else:
            self.model = model
            self._model_bound = True

    @property
    def is_bound(self) -> bool:
        """Returns true if the Filter has been bound to a Model"""
        return self._model_bound

    def get_field(self, key: str) -> Any:
        """
        Gets a field by name.

        :param key: The name of the field.
        :type key: str
        :return: The value of the field.
        :rtype: Any
        :raises KeyError: if the requested field does not exist.
        """
        return self._fields[key]

    def set_field(self, key: str, value: Any) -> None:
        """
        Sets the value of a field by name

        :param key: The name of the field.
        :type key: str
        :param value: The value of the field.
        :type value: Any
        """
        self._fields[key] = value

    @property
    def field_names(self) -> list[str]:
        """The list of field names."""
        return list(self._fields.keys())

    @classmethod
    def from_conf(cls, name: str, conf: dict[str, Any], default: dict[str, Any] | None = None) -> Self:
        """
        Builds a Filter object from a steering file dictionary.

        If the name is in dotted notation, then this should be corresponding to the table in the configuration file.
        If a default configuration is provided, this will be used as a starting point for the filter, and it will be
        updated by the actual configuration in ``conf``.

        In normal use, you would provide the specific configuration via the conf parameter and the global filter
        configuration as default.

        See details in the :class:`class documentation <Filter>`

        :param name: The name of the filter in dotted notation.
        :type name: str
        :param conf: The configuration dictionary.
        :type conf: dict
        :param default: Default configuration dictionary
        :type default: dict
        :return: A Filter object
        :rtype: Filter
        """
        param = default or {}

        # split the name from dotted notation
        # ProcessorName.ModelName.Filter
        names = name.split('.')
        if len(names) == 3 and names[1] == 'Filter':
            proc_name, _, model_name = names
            if proc_name in conf and 'Filter' in conf[proc_name] and model_name in conf[proc_name]['Filter']:
                param.update(copy(conf[proc_name]['Filter'][model_name]))

        # if the name is not in the expected dotted notation, then param will be the default, that very likely means
        # the global filter configuration.
        return cls(name, **param)

    def filter(self) -> peewee.Expression | bool:
        """
        Generates a filtering expression logically ANDing all filtering fields.

        See details in the :class:`class documentation <Filter>`

        :return: The filtering expression.
        :rtype: peewee.Expression
        :raises TypeError: when the field value type is not supported.
        """
        if not self.is_bound:
            warnings.warn('Unable to generate the filter. Did you bind the filter to the model?')
            return True
        else:
            expression_list = []
            for field, value in self._fields.items():
                if hasattr(self.model, field):
                    if isinstance(value, (int, float, bool)):
                        expression_list.append(getattr(self.model, field) == value)
                    elif isinstance(value, str):
                        expression_list.append(getattr(self.model, field) % value)
                    elif isinstance(value, list):
                        expression_list.append(getattr(self.model, field).in_(value))
                    else:
                        raise TypeError(f'Filter value of unsupported type {type(value)}.')
            return reduce(operator.and_, expression_list, True)


class FilterRegister(UserDict[str, Filter]):
    """
    A special dictionary to store all :class:`Filters <mafw.db.db_filter.Filter>` in a processors.

    It contains a publicly accessible dictionary with the configuration of each Filter using the Model name as
    keyword.

    It contains a private dictionary with the global filter configuration as well.
    The global filter is not directly accessible, but only some of its members will be exposed via properties.
    In particular, the new_only flag that is relevant only at the Processor level can be accessed directly using the
    :attr:`new_only`. If not specified in the configuration file, the new_only is by default True.
    """

    def __init__(self, data: dict[str, Filter] | None = None, /, **kwargs: Any) -> None:
        """
        Constructor parameters:

        :param data: Initial data
        :type data: dict
        :param kwargs: Keywords arguments
        """
        self._global_filter: dict[str, Any] = {}
        super().__init__(data, **kwargs)

    @property
    def new_only(self) -> bool:
        """
        The new only flag.

        :return: True, if only new items, not already in the output database table must be processed.
        :rtype: bool
        """
        return cast(bool, self._global_filter.get('new_only', True))

    @new_only.setter
    def new_only(self, v: bool) -> None:
        self._global_filter['new_only'] = v

    def __setitem__(self, key: str, value: Filter) -> None:
        """
        Set a new value at key.

        If value is not a Filter, then it will be automatically and silently discarded.

        :param key: Dictionary key. Normally the name of the model linked to the filter.
        :type key: str
        :param value: The Filter.
        :type value: Filter
        """
        if not isinstance(value, Filter):
            return
        super().__setitem__(key, value)

    def bind_all(self, models: list[type[Model]] | dict[str, type[Model]]) -> None:
        """
        Binds all filters to their models.

        The ``models`` list or dictionary should contain a valid model for all the Filter in the registry.
        In the case of a dictionary, the key value should be the model name.

        If the user provides a model for which there is no corresponding filter in the register, then a new filter
        for that model is created using the GlobalFilter default.

        This can happen if the user did not provide a specific table for the Processor/Model, but simply put all
        filtering conditions in the GlobalFilter table. Even though, this behaviour is allowed and working,
        it may result in unexpected results. Also listing more than needed models in the input list can be dangerous
        because they will anyhow use the default filters.

        :param models: List or dictionary of a databank of Models from which the Filter can be bound.
        :type models:  list[type(Model)] | dict[str,type(Model)]
        """
        if isinstance(models, list):
            models = {m.__name__: m for m in models}

        # check, if we have a filter for each listed models, if not create one using the default configuration.
        for model_name in models.keys():
            if model_name not in self.data:
                self.data[model_name] = Filter.from_conf(f'{model_name}', conf={}, default=self._global_filter)

        for k, v in self.data.items():
            if k in self.data and k in models:
                v.bind(models[k])
            else:
                v.bind()

    def filter_all(self):  # type: ignore[no-untyped-def]
        """
        Generates a where clause ANDing all filters.

        If one Filter is not bound, then True is returned.
        """
        filter_list = [flt.filter() for flt in self.data.values() if flt.is_bound]
        return reduce(operator.and_, filter_list, True)



#  Copyright 2025 European Union
#  Author: Bulgheroni Antonio (antonio.bulgheroni@ec.europa.eu)
#  SPDX-License-Identifier: EUPL-1.2
"""
The module provides some examples for the user to develop their own processors.

Those implemented here are mainly used in the test suit.
"""

from mafw.enumerators import LoopType
from mafw.processor import ActiveParameter, Processor


class AccumulatorProcessor(Processor):
    r"""
    A processor to calculate the sum of the first n values via a looping approach.

    In mathematical terms, this processor solves this easy equation:

    .. math::

        N = \sum_{i=0}^{n}{i}

    by looping. It is a terribly inefficient approach, but it works as a demonstration of the looping structure.

    The user can get the results by retrieving the `accumulated_value` parameter at the end of the processor
    execution.
    """

    last_value = ActiveParameter('last_value', default=100, help_doc='Last value of the series')

    def __init__(self, *args, **kwargs):
        """Constructor parameters:

        :param last_value: The `n` in the equation above. Defaults to 100
        :type last_value: int
        :param accumulated_value: The `N` in the equation above at the end of the process.
        :type accumulated_value: int
        """
        super().__init__(*args, **kwargs)
        self.accumulated_value: int = 0

    def start(self):
        """Resets the accumulated value to 0 before starting."""
        super().start()
        self.accumulated_value = 0

    def get_items(self) -> list[int]:
        """Returns the list of the first `last_value` integers."""
        return list(range(self.last_value))

    def process(self):
        """Increase the accumulated value by the current item."""
        self.accumulated_value += self.item


class GaussAdder(Processor):
    r"""
    A processor to calculate the sum of the first n values via the so called *Gauss formula*.

    In mathematical terms, this processor solves this easy equation:

    .. math::

        N = \frac{n * (n - 1)}{2}

    without any looping

    The user can get the results by retrieving the `sum_value` parameter at the end of the processor
    execution.
    """

    last_value = ActiveParameter('last_value', default=100, help_doc='Last value of the series.')

    def __init__(self, *args, **kwargs):
        """
        Constructor parameters:

        :param last_value: The `n` in the equation above. Defaults to 100
        :type last_value: int
        :param sum_value: The `N` in the equation above.
        :type sum_value: int
        """
        super().__init__(looper=LoopType.SingleLoop, *args, **kwargs)
        self.sum_value: int = 0

    def start(self):
        """Sets the sum value to 0."""
        super().start()
        self.sum_value = 0

    def process(self):
        """Compute the sum using the Gauss formula."""
        self.sum_value = int(self.last_value * (self.last_value - 1) / 2)
